from io import StringIO
from typing import Any

# Tests classes
import pytest

from freeorion_tools.chat_handler.chat_formatter import ChatFormatter
from freeorion_tools.chat_handler.debug_chat_handler import DebugChatHandler
from freeorion_tools.chat_handler.shell_variable import ShellVariable


class Printer:
    def __init__(self):
        self.input = StringIO()

    def __call__(self, message):
        self.input.write(str(message))
        self.input.write("\n")


class DummyProvider:
    def __init__(self, player_id):
        self._player_id = player_id

    def get_player_ids(self):
        return [1, 2, 3]

    def is_host(self, player_id):
        return player_id == 1

    def player_id(self):
        return self._player_id

    def player_name(self, player_id):
        return "Player_%s" % player_id

    def get_empire_id(self, player_id):
        return player_id

    def get_empire_name(self, player_id):
        return "Empire_%s" % player_id

    def get_empire_color(self, player_id):
        return 0, 0, 0, 0


class DummyFormatter(ChatFormatter):
    def _get_tag_rgb(self, color: (int, int, int, int)) -> (str, str):
        return "", ""

    def _get_underline_tag(self) -> (str, str):
        return "", ""

    def _wrap_to_tag(self, tag: str, message: Any) -> str:
        return str(message)


@pytest.fixture
def debug_handler():
    _shell_locals = [
        ShellVariable(
            variable="variable_a",
            expression="'value_a'",
            description="description for a",
        ),
        ShellVariable(
            variable="variable_b",
            expression="'value_b'",
            description="description for b",
        ),
        ShellVariable(
            variable="variable_c",
            expression="'value_c'",
            description="variable with import",
            imports=("from glob import has_magic",),
        ),
        ShellVariable(
            variable="variable_a_plus_b",
            expression="variable_a + variable_b",
            description="description for a+b",
        ),
    ]

    provider = DummyProvider(2)
    return DebugChatHandler(
        2,
        provider=provider,
        shell_locals=_shell_locals,
        formatter=DummyFormatter(),
        chat_printer=Printer(),
        log_printer=Printer(),
    )


@pytest.fixture
def debug_handler_for_muted_ai():
    provider = DummyProvider(3)
    chat_printer = Printer()
    return chat_printer.input, DebugChatHandler(
        3,
        provider=provider,
        shell_locals=[],
        formatter=DummyFormatter(),
        chat_printer=chat_printer,
        log_printer=Printer(),
    )


def get_output_from_handler(handler):
    chat = handler._chat_printer.input.getvalue()
    log = handler._log_printer.input.getvalue()
    return chat, log


@pytest.fixture
def test_caller(debug_handler: DebugChatHandler):
    def function(message):
        # clear previous logs
        debug_handler._chat_printer.input.truncate(0)
        debug_handler._log_printer.input.truncate(0)
        assert debug_handler.process_message(1, message)
        chat, log = get_output_from_handler(debug_handler)
        assert "SyntaxError" not in chat
        assert "SyntaxError" not in log
        return chat, log

    return function


class TestNonDebugModeForAiWhoResponds:
    def test_send_help_prints_user_names(self, test_caller):
        chat, log = test_caller("help")
        assert "player_1" not in chat
        assert "Empire_2 by Player_2" in chat
        assert "Empire_3 by Player_3" in chat

    def test_sending_unknown_message_is_ignored_by_handler(self, debug_handler: DebugChatHandler):
        handled = debug_handler.process_message(1, "hello world")
        assert handled is False
        chat, log = get_output_from_handler(debug_handler)
        assert chat == ""
        assert log == ""

    def test_sending_stop_before_start_is_swallowed_by_handler(self, debug_handler: DebugChatHandler):
        result = debug_handler.process_message(1, "stop")
        assert result is True
        chat, log = get_output_from_handler(debug_handler)
        assert chat == ""
        assert log == ""


class TestNonDebugModeForAiWhoMuted:
    def test_send_help_is_ignored_by_second_ai(self, debug_handler_for_muted_ai: tuple[StringIO, DebugChatHandler]):
        chat_io, handler = debug_handler_for_muted_ai
        assert handler.process_message(1, "help")
        assert chat_io.getvalue() == ""


class TestDebugModStart:
    def test_sending_message_from_ai_is_propagated_by_handler(self, debug_handler: DebugChatHandler):
        # ignore messages not from human
        result = debug_handler.process_message(3, "start 2")
        assert result is False
        chat, log = get_output_from_handler(debug_handler)
        assert chat == ""
        assert log == ""

    def test_starting_with_not_existing_ai_is_swallowed_by_handler(self, debug_handler: DebugChatHandler):
        result = debug_handler.process_message(1, "start 10")
        assert result is True
        chat, log = get_output_from_handler(debug_handler)
        assert chat == ""
        assert log == ""

    def test_starting_with_nonnumerical_ai_is_swallowed_by_handler(self, debug_handler: DebugChatHandler):
        assert debug_handler.process_message(1, "start xxx")
        chat, log = get_output_from_handler(debug_handler)
        assert "Invalid empire id, please input valid number" in chat

    def test_start_start_debug_mode_for_respond_ai(self, test_caller):
        chat, log = test_caller("start 2")
        assert "Entering debug mode" in chat

    def test_start_start_debug_mode_for_muted_ai(self, debug_handler_for_muted_ai: tuple[StringIO, DebugChatHandler]):
        chat_io, handler = debug_handler_for_muted_ai
        handler.process_message(1, "start 3")
        assert "Entering debug mode" in chat_io.getvalue()

    def test_sending_start_prints_intro_message(self, test_caller):
        chat, log = test_caller("start 2")
        assert "variable_a" in chat
        assert "description for a" in chat
        assert "variable_b" in chat
        assert "description for b" in chat
        assert "variable_a_plus_b" in chat
        assert "scription for a+b" in chat


@pytest.fixture
def start_debug(test_caller):
    return test_caller("start 2")


@pytest.mark.usefixtures("start_debug")
class TestInDebugMode:
    def test_sending_stop_pauses_debug(self, test_caller):
        chat, log = test_caller("stop")
        assert "Exiting debug mode" in chat

    def test_evaluate_initial_variables_prints_them_to_chat(self, test_caller):
        chat, log = test_caller("variable_a")
        assert "value_a" in chat
        chat, log = test_caller("variable_b")
        assert "value_b" in chat
        chat, log = test_caller("variable_a_plus_b")
        assert "value_avalue_b" in chat

    def test_evaluating_command_prints_command_to_log(self, test_caller):
        chat, log = test_caller("'free' + 'orion'")
        assert "'free' + 'orion'" in log

    def test_evaluating_command_prints_result(self, test_caller):
        chat, log = test_caller("'free' + 'orion'")
        assert "freeorion" in chat

    def test_evaluating_assignment_assigns_variable(self, test_caller):
        test_caller("variable_c = 'value_c'")
        chat, log = test_caller("variable_c")
        assert "value_c" in chat

    def test_evaluating_erroneous_statement_prints_exception_to_console(self, debug_handler: DebugChatHandler):
        assert debug_handler.process_message(1, "1/0")
        chat, log = get_output_from_handler(debug_handler)
        assert "ZeroDivisionError: division by zero" in chat
        assert "ZeroDivisionError: division by zero" in log
        assert debug_handler.process_message(1, "variable_a")
        chat, log = get_output_from_handler(debug_handler)
        assert "value_a" in chat

    def test_start_with_initial_imports_evaluates_imports(self, test_caller):
        chat, log = test_caller("has_magic")
        assert "function has_magic at" in chat

    def test_help_call_is_ignored(self, test_caller):
        chat, log = test_caller("help")
        assert "Type help() for interactive help" in chat

    def test_start_call_is_ignored(self, test_caller):
        chat, log = test_caller("start")
        assert " name 'start' is not defined" in chat

    def test_start_with_number_call_is_ignored(self, debug_handler: DebugChatHandler):
        assert debug_handler.process_message(1, "start 2")
        chat, log = get_output_from_handler(debug_handler)
        assert "SyntaxError" in log
        assert "SyntaxError" in chat


class TestRestart:
    def test_modified_initial_variables_are_restored(self, test_caller):
        test_caller("start 2")
        test_caller("variable_b = 'value_d'")
        chat, log = test_caller("variable_b")
        assert "value_d" in chat
        test_caller("stop")
        test_caller("start 2")
        chat, log = test_caller("variable_b")
        assert "value_b" in chat
        assert "value_d" not in chat

    def test_not_initial_variables_are_preserved(self, test_caller):
        test_caller("start 2")
        test_caller("variable_d = 'value_d'")
        chat, log = test_caller("variable_d")
        assert "value_d" in chat
        test_caller("stop")
        test_caller("start 2")
        chat, log = test_caller("variable_d")
        assert "value_d" in chat



from focs._effects import (
    AddSpecial,
    Conditional,
    ContainedBy,
    Contains,
    EffectsGroup,
    GalaxyMaxAIAggression,
    HasSpecial,
    HasSpecies,
    HasTag,
    IsSource,
    MaxOf,
    OneOf,
    Planet,
    Random,
    Source,
    Turn,
)
from macros.misc import PLANET_DEFENSE_FACTOR, PLANET_SHIELD_FACTOR
from species.species_macros.detection import NATIVE_PLANETARY_DETECTION
from species.species_macros.planet_defense import AVERAGE_PLANETARY_DEFENSE, NATIVE_PLANETARY_DEFENSE
from species.species_macros.planet_shields import AVERAGE_PLANETARY_SHIELDS, _native_planetary_shields

_FORTIFICATION_PROBABILITY = 0.1 * MaxOf(float, 0.2 * Source.TargetPopulation, Source.Research + Source.Industry)


DEFAULT_NATIVE_DEFENSE = [
    *AVERAGE_PLANETARY_SHIELDS,
    *AVERAGE_PLANETARY_DEFENSE,
    EffectsGroup(
        scope=IsSource & Turn(low=1, high=1) & (GalaxyMaxAIAggression >= 1),
        activation=Planet()
        & Random(probability=_FORTIFICATION_PROBABILITY)
        & ~ContainedBy(
            Contains(
                Planet() & HasSpecial(name="MODERATE_TECH_NATIVES_SPECIAL")
                | Planet() & HasSpecial(name="HIGH_TECH_NATIVES_SPECIAL")
                | HasSpecies() & HasTag(name="PRIMITIVE")
            )
        ),
        effects=[
            Conditional(
                condition=~Random(probability=_FORTIFICATION_PROBABILITY),
                effects=[
                    AddSpecial(
                        name=OneOf(
                            str,
                            "NATIVE_FORTIFICATION_MINIMAL",
                            "NATIVE_FORTIFICATION_LOW",
                            "CLOUD_COVER_MASTER_SPECIAL",
                        )
                    ),
                ],
                else_=[
                    AddSpecial(
                        name=OneOf(
                            str,
                            "NATIVE_FORTIFICATION_MEDIUM",
                            "NATIVE_FORTIFICATION_HIGH",
                            "VOLCANIC_ASH_MASTER_SPECIAL",
                        )
                    ),
                ],
            )
        ],
    ),
]

ADVANCED_NATIVE_DEFENSE = [
    *AVERAGE_PLANETARY_SHIELDS,
    *AVERAGE_PLANETARY_DEFENSE,
    NATIVE_PLANETARY_DETECTION(10),
    NATIVE_PLANETARY_DEFENSE(10 * PLANET_DEFENSE_FACTOR),
    _native_planetary_shields(10 * PLANET_SHIELD_FACTOR),
    EffectsGroup(
        scope=IsSource & Turn(low=1, high=1) & (GalaxyMaxAIAggression >= 1),
        activation=Planet(),
        effects=[
            Conditional(
                condition=~ContainedBy(
                    Contains(
                        Planet() & HasSpecial(name="MODERATE_TECH_NATIVES_SPECIAL")
                        | Planet() & HasSpecial(name="HIGH_TECH_NATIVES_SPECIAL")
                    ),
                ),
                effects=[
                    AddSpecial(
                        name=OneOf(
                            str,
                            "NATIVE_FORTIFICATION_LOW",
                            "NATIVE_FORTIFICATION_MEDIUM",
                            "CLOUD_COVER_MASTER_SPECIAL",
                            "VOLCANIC_ASH_MASTER_SPECIAL",
                        )
                    )
                ],
                else_=[],
            )
        ],
    ),
]



from buildings.buildings import SPECIES_LIKES_OR_DISLIKES_BUILDING_STABILITY_EFFECTS
from focs._effects import (
    Contains,
    HasSpecial,
    IsBuilding,
    OwnedBy,
    Planet,
    Source,
)
from macros.base_prod import BUILDING_COST_MULTIPLIER
from macros.enqueue import ENQUEUE_BUILD_ONE_PER_PLANET

try:
    from focs._buildings import *
except ModuleNotFoundError:
    pass

BuildingType(  # type: ignore[reportUnboundVariable]
    name="BLD_BIOTERROR_PROJECTOR",
    description="BLD_BIOTERROR_PROJECTOR_DESC",
    buildcost=75 * BUILDING_COST_MULTIPLIER,
    buildtime=8,
    location=(
        Planet()
        & ~Contains(IsBuilding(name=["BLD_BIOTERROR_PROJECTOR"]))
        & OwnedBy(empire=Source.Owner)
        & HasSpecial(name="RESONANT_MOON_SPECIAL")
        & ~Contains(IsBuilding(name=["BLD_BIOTERROR_PROJECTOR"]))
    ),
    enqueuelocation=ENQUEUE_BUILD_ONE_PER_PLANET,
    effectsgroups=[
        *SPECIES_LIKES_OR_DISLIKES_BUILDING_STABILITY_EFFECTS,
    ],
    icon="icons/building/bioterror_projector.png",
)



from focs._species import *
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import INFERNO_NARROW_EP, NARROW_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import BAD_HAPPINESS
from species.species_macros.industry import GREAT_INDUSTRY
from species.species_macros.influence import BAD_INFLUENCE
from species.species_macros.planet_defense import AVERAGE_PLANETARY_DEFENSE
from species.species_macros.planet_shields import AVERAGE_PLANETARY_SHIELDS
from species.species_macros.planet_size import LARGE_PLANET
from species.species_macros.population import BAD_POPULATION
from species.species_macros.research import BAD_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import GREAT_SUPPLY
from species.species_macros.troops import GREAT_DEFENSE_TROOPS, GREAT_OFFENSE_TROOPS

Species(
    name="SP_EGASSEM",
    description="SP_EGASSEM_DESC",
    gameplay_description="SP_EGASSEM_GAMEPLAY_DESC",
    playable=True,
    can_produce_ships=True,
    can_colonize=True,
    tags=[
        "LITHIC",
        "ARTISTIC",
        "BAD_RESEARCH",
        "GREAT_INDUSTRY",
        "BAD_INFLUENCE",
        "BAD_POPULATION",
        "BAD_HAPPINESS",
        "GREAT_SUPPLY",
        "GREAT_OFFENSE_TROOPS",
        "PEDIA_LITHIC_SPECIES_CLASS",
        "PEDIA_ARTISTIC",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_INDUSTRY",
    likes=[
        "FOCUS_INDUSTRY",
        "FRACTAL_GEODES_SPECIAL",
        "SHIMMER_SILK_SPECIAL",
        "PROBIOTIC_SPECIAL",
        "MONOPOLE_SPECIAL",
        "ELERIUM_SPECIAL",
        "FORTRESS_SPECIAL",
        "PLC_CONFEDERATION",
        "PLC_CAPITAL_MARKETS",
        "PLC_CONFORMANCE",
        "PLC_FEUDALISM",
        "PLC_ALGORITHMIC_RESEARCH",
        "PLC_DESIGN_SIMPLICITY",
        "PLC_MARINE_RECRUITMENT",
        "PLC_MARTIAL_LAW",
        "PLC_PROPAGANDA",
    ],
    dislikes=[
        "BLD_NEUTRONIUM_EXTRACTOR",
        "BLD_BLACK_HOLE_POW_GEN",
        "BLD_STARGATE",
        "BLD_SHIPYARD_AST_REF",
        "BLD_SHIPYARD_AST",
        "HONEYCOMB_SPECIAL",
        "WORLDTREE_SPECIAL",
        "KRAKEN_IN_THE_ICE_SPECIAL",
        "KRAKEN_NEST_SPECIAL",
        "PLC_ENVIRONMENTALISM",
        "PLC_BOOTSTRAPPING",
        "PLC_CENTRALIZATION",
        "PLC_LIBERTY",
        "PLC_NO_GROWTH",
        "PLC_DREAM_RECURSION",
        "PLC_ENGINEERING",
    ],
    effectsgroups=[
        *GREAT_INDUSTRY,
        *BAD_RESEARCH,
        *BAD_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *BAD_POPULATION,
        *BAD_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_EGASSEM"),
        *GREAT_SUPPLY,
        *GREAT_DEFENSE_TROOPS,
        *GREAT_OFFENSE_TROOPS,
        # not for description,
        *AVERAGE_PLANETARY_SHIELDS,
        *AVERAGE_PLANETARY_DEFENSE,
        *LARGE_PLANET,
        NARROW_EP,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=INFERNO_NARROW_EP,
    graphic="icons/species/egassem.png",
)



"""
ResourcesAI.py provides generate_resources_orders which sets the focus for all populated planets in the empire.

The method is to start with a raw list of all the populated planets in the empire.
It considers in turn growth factors, production specials, defense requirements
and finally the targeted ratio of research/production. Each decision on a planet
transfers the planet from the raw list to the baked list, until all planets
have their future focus decided.
"""
from __future__ import annotations

import freeOrionAIInterface as fo
from collections.abc import Sequence
from functools import total_ordering
from itertools import chain
from logging import debug, info, warning
from operator import itemgetter
from typing import NamedTuple

import AIDependencies
import PlanetUtilsAI
import ProductionAI
from aistate_interface import get_aistate
from buildings import BuildingType
from colonization.calculate_planet_colonization_rating import empire_metabolisms
from common.fo_typing import PlanetId
from common.print_utils import Table, Text
from empire.growth_specials import get_growth_specials
from EnumsAI import FocusType, PriorityType, get_priority_resource_types
from freeorion_tools import get_named_real, get_species_industry, get_species_research, tech_is_complete
from freeorion_tools.bonus_calculation import adjust_direction
from freeorion_tools.statistics import stats
from freeorion_tools.timers import AITimer
from PolicyAI import PolicyManager, bureaucracy, liberty
from turn_state import (
    computronium_candidates,
    honeycomb_candidates,
    population_with_industry_focus,
)

resource_timer = AITimer("timer_bucket")

# Local Constants
INDUSTRY = FocusType.FOCUS_INDUSTRY
RESEARCH = FocusType.FOCUS_RESEARCH
GROWTH = FocusType.FOCUS_GROWTH
PROTECTION = FocusType.FOCUS_PROTECTION
INFLUENCE = FocusType.FOCUS_INFLUENCE
supported_foci = {INDUSTRY, RESEARCH, INFLUENCE, GROWTH, PROTECTION}
USELESS_RATING = -99999  # focus is either not available or stability is too bad


def _focus_name(focus: str) -> str:
    _known_names = {
        INDUSTRY: "Industry",
        RESEARCH: "Research",
        GROWTH: "Growth",
        PROTECTION: "Defense",
        INFLUENCE: "Influence",
    }
    return _known_names.get(focus, focus)


class Output(NamedTuple):
    industry: float
    research: float
    influence: float
    stability: float
    rating: float  # sum of all production values weighed by priorities for overall comparisons


class FocusRating(NamedTuple):
    rating: float
    focus: str


class PlanetFocusRating:
    """
    Best and second after the best focuses for the planet.
    """

    def __init__(self, focuses: Sequence[tuple[float, str]]):
        """
        Initialize object wih best and second best focuses.

        There is a number of focuses equal length of the `supported_foci`
        """
        assert len(focuses) == len(supported_foci)
        focuses = sorted((FocusRating(rating, focus) for rating, focus in focuses), reverse=True)
        self.best = focuses[0]
        self.second = focuses[1]


@total_ordering
class PlanetFocusInfo:
    """The current, possible and future foci and output of one planet."""

    def __init__(self, planet: fo.planet):
        self.planet = planet
        # Make sure current_focus is set, without it some calculation may go astray.
        self.current_focus = planet.focus if planet.focus else fo.getSpecies(planet.speciesName).preferredFocus
        self.current_output = Output(
            industry=planet.currentMeterValue(fo.meterType.industry),
            research=planet.currentMeterValue(fo.meterType.research),
            influence=planet.currentMeterValue(fo.meterType.influence),
            stability=planet.currentMeterValue(fo.meterType.happiness),
            rating=0.0,  # cannot calculate the rating here, do we even need it?
        )
        self.possible_output = {}
        self.future_focus = self.current_focus
        self.rated_foci: PlanetFocusRating = None  # initialized after the object creation
        self.options = {focus for focus in planet.availableFoci if focus in supported_foci}

    def __repr__(self):
        return f"PlanetFocusInfo({repr(self.planet)})"

    def __eq__(self, other: PlanetFocusInfo):
        return self.planet.id == other.planet.id

    def __lt__(self, other: PlanetFocusInfo):
        """
        Dummy sorting.
        In this module, we sort Tuples[float, PlanetFocusInfo]. To do so, PlanetFocusInfo must be sortable, but if the
        float is the same, we don't really care which comes first.
        """
        return self.planet.id < other.planet.id

    def set_rated_foci(self):
        self.rated_foci = PlanetFocusRating([(self.possible_output[focus].rating, focus) for focus in supported_foci])

    def rating_below_best(self, focus: str) -> float:
        return self.possible_output[focus].rating - self.rated_foci.best.rating

    def best_over_second(self) -> float:
        return self.rated_foci.best.rating - self.rated_foci.second.rating

    def difference(self, focus1: str, focus2: str) -> float:
        """How much is the output of focus1 better or worse than output of focus2."""
        return self.possible_output[focus1].rating - self.possible_output[focus2].rating

    def evaluate(self, focus: str) -> float:
        """
        Compare focus with best alternative.
        If focus produces best, return a positive rating, giving how much its output is better than second best.
        If another focus produces better, returns a negative rating, giving how much it is worse than best.
        Result is divided by sqrt(population) to prefer switching smaller planets.
        """
        sqrt_population = max(1.0, self.planet.currentMeterValue(fo.meterType.population)) ** 0.5
        if sqrt_population == 0:
            sqrt_population = 0.0001  # prevent Zero division

        if focus == self.rated_foci.best.focus:
            # TBD replace division by using immediate_lost...
            return self.best_over_second() / sqrt_population
        else:
            return self.rating_below_best(focus) / sqrt_population


class PlanetFocusManager:
    """PlanetFocusManager tracks all of the empire's planets, what their current and future focus will be."""

    def __init__(self):
        universe = fo.getUniverse()

        resource_timer.start("getPlanets")
        self.planet_ids = [planet.id for planet in PlanetUtilsAI.get_empire_populated_planets()]

        resource_timer.start("Targets")
        self.planet_info: dict[PlanetId, PlanetFocusInfo] = {
            pid: PlanetFocusInfo(universe.getPlanet(pid)) for pid in self.planet_ids
        }
        self.baked_planet_info = {}
        aistate = get_aistate()
        self.priority_industry = aistate.get_priority(PriorityType.RESOURCE_PRODUCTION)
        self.priority_research = aistate.get_priority(PriorityType.RESOURCE_RESEARCH)
        self.priority_influence = aistate.get_priority(PriorityType.RESOURCE_INFLUENCE)
        self.calculate_planet_infos()

    def set_planetary_foci(self, reporter: Reporter):
        self.set_planet_growth_specials()
        self.set_planet_production_and_research_specials()
        reporter.capture_section_info("Specials")
        # TODO redo. Note that setting protection for stability is handled production_value and set_influence
        # self.set_planet_protection_foci()
        # reporter.capture_section_info("Protection")
        self.set_influence_focus()
        reporter.capture_section_info("Influence")
        self.early_capital_handling()
        self.set_other_foci()
        reporter.capture_section_info("Typical")
        reporter.print_table(self.priority_research / self.priority_industry)

    def bake_future_focus(self, pid, focus, force=False):
        """Set the focus and moves it from the raw list to the baked list of planets.
        pid -- pid
        focus -- future focus to use
        Return success or failure
        """
        pinfo = self.planet_info.get(pid)
        debug(f"baking {focus} on {pinfo.planet}")
        if not pinfo:
            return False
        if (focus == INDUSTRY or focus == RESEARCH) and not force:
            if self.check_avoid_change_due_to_bureaucracy(pinfo, focus):
                return False

        success = bool(
            pinfo.current_focus == focus
            or (focus in pinfo.planet.availableFoci and fo.issueChangeFocusOrder(pid, focus))
        )
        if success:
            if pinfo.current_focus != focus:
                PolicyManager.report_focus_change()
            pinfo.future_focus = focus
            self.baked_planet_info[pid] = self.planet_info.pop(pid)
        debug(f"success={success}")
        return success

    def check_avoid_change_due_to_bureaucracy(self, pinfo: PlanetFocusInfo, focus: str) -> bool:
        """Check whether we better not change a focus to avoid the influence penalty from bureaucracy."""
        if bureaucracy not in fo.getEmpire().adoptedPolicies or focus == pinfo.current_focus:
            return False  # No bureaucracy or no change
        last_turn = fo.currentTurn() - 1
        if pinfo.planet.lastTurnColonized == last_turn or pinfo.planet.lastTurnConquered == last_turn:
            return False  # no penalty for newly settled/conquered planets
        if pinfo.current_output.influence + 0.5 < pinfo.possible_output[pinfo.current_focus].influence:
            # avoid repeated focus changes: When influence production already is below target, do not
            # get it further down by switching again (I've seen one planet at -30...)
            debug(f"Avoid focus change on {pinfo.planet.name} due bureaucracy and influence not maxed.")
            return True
        return False

    def rate_output(self, pp: float, rp: float, ip: float, stability: float, focus: str) -> float:
        """Calculate rating of the given production output."""
        # For some reason even rebelling planets can produce influence. Also, we must avoid a death cycle: when
        # stability goes down due to influence debt, trying to counter instability by setting influence producing
        # planets to protection focus just makes it worse.
        if stability <= 0.0 and focus != INFLUENCE:
            return USELESS_RATING
        return (
            pp * self.priority_industry
            + rp * self.priority_research
            + ip * self.priority_influence
            # add a small amount for stability to make sure output of protection is better than growth
            + stability / 10
        )

    def possible_output_with_focus(self, planet: fo.planet, focus: str, stability: float | None = None) -> Output:
        """
        Estimate Output of planet, which should have been set to focus and meters should be current.
        Special case: if stability is giving, calculate output from current focus, but with adapted stability.
        """
        if planet.focus != focus and stability is None:  # basic output of growth is determined using protection
            return Output(0.0, 0.0, 0.0, 0.0, USELESS_RATING)

        industry_target = planet.currentMeterValue(fo.meterType.targetIndustry)
        research_target = planet.currentMeterValue(fo.meterType.targetResearch)
        influence_target = planet.currentMeterValue(fo.meterType.targetInfluence)
        current_stability = planet.currentMeterValue(fo.meterType.happiness)
        # note that stability 0.0 is a valid value
        target_stability = planet.currentMeterValue(fo.meterType.targetHappiness) if stability is None else stability
        if liberty in fo.getEmpire().adoptedPolicies and focus == RESEARCH:
            research_target += PlanetUtilsAI.adjust_liberty(planet, planet.currentMeterValue(fo.meterType.population))
        if tech_is_complete(AIDependencies.PRO_AUTO_1):
            min_stability = get_named_real("PRO_ADAPTIVE_AUTO_MIN_STABILITY")
            flat = get_named_real("PRO_ADAPTIVE_AUTO_TARGET_INDUSTRY_FLAT")
            industry_target += flat * adjust_direction(min_stability, current_stability, target_stability)
        if tech_is_complete(AIDependencies.LRN_ARTIF_MINDS_1):
            min_stability = get_named_real("LRN_NASCENT_AI_MIN_STABILITY")
            flat = get_named_real("LRN_NASCENT_AI_TARGET_RESEARCH_FLAT")
            research_target += flat * adjust_direction(min_stability, current_stability, target_stability)
        if target_stability < 0:
            industry_target = research_target = 0.0
        # There are a lot more adjustments, some also depend on supply connection. Hopefully one day we get
        # meter_update with adjusted stability...

        return Output(
            industry=industry_target,
            research=research_target,
            influence=influence_target,
            stability=target_stability,
            rating=self.rate_output(industry_target, research_target, influence_target, target_stability, focus),
        )

    def calculate_planet_infos(self):  # noqa complexity
        """
        Calculates for each possible focus an estimated target output of each planet and stores it in planet info.
        """

        def set_focus(pinfo, focus):
            if focus in pinfo.planet.availableFoci and pinfo.planet.focus != focus:
                fo.issueChangeFocusOrder(pinfo.planet.id, focus)

        def collect_estimates(planet_info, focus, callback, second_focus=None):
            for pinfo in planet_info.values():
                set_focus(pinfo, focus)
            universe.updateMeterEstimates(list(planet_info))
            for pinfo in planet_info.values():
                pinfo.possible_output[focus] = callback(pinfo.planet, focus)
            if second_focus:
                for pinfo in planet_info.values():
                    stability = PlanetUtilsAI.stability_with_focus(pinfo.planet, GROWTH)
                    pinfo.possible_output[GROWTH] = callback(pinfo.planet, PROTECTION, stability)

        universe = fo.getUniverse()
        collect_estimates(self.planet_info, INDUSTRY, self.possible_output_with_focus)
        collect_estimates(self.planet_info, RESEARCH, self.possible_output_with_focus)
        collect_estimates(self.planet_info, INFLUENCE, self.possible_output_with_focus)
        # PROTECTION and GROWTH only produce through focus-less bonuses, so they can do with one meter estimation
        collect_estimates(self.planet_info, PROTECTION, self.possible_output_with_focus, GROWTH)
        # revert to current focus and update meters again
        for pinfo in self.planet_info.values():
            set_focus(pinfo, pinfo.current_focus)
        universe.updateMeterEstimates(self.planet_ids)

        for pinfo in self.planet_info.values():
            pinfo.set_rated_foci()
        self.print_pinfo_table()

    def print_pinfo_table(self):
        def output_table_format(o: Output) -> str:
            if o.rating == USELESS_RATING:
                return "---"
            else:
                return f"{o.rating:.1f} {o.industry:.1f} {o.research:.1f} {o.influence:.1f} {o.stability:.1f}"

        debug("Values per focus: rating pp rp ip stability")
        pinfo_table = Table(
            Text("Planet"),
            Text("Best Focus"),
            Text("Industry"),
            Text("Research"),
            Text("Influence"),
            Text("Protection"),
            Text("Growth"),
            table_name="Potential Planetary Output Overview Turn %d" % fo.currentTurn(),
        )
        for pinfo in self.planet_info.values():
            pinfo_table.add_row(
                pinfo.planet,
                _focus_name(pinfo.rated_foci.best.focus),
                output_table_format(pinfo.possible_output[INDUSTRY]),
                output_table_format(pinfo.possible_output[RESEARCH]),
                output_table_format(pinfo.possible_output[INFLUENCE]),
                output_table_format(pinfo.possible_output[PROTECTION]),
                output_table_format(pinfo.possible_output[GROWTH]),
            )
        info(pinfo_table)

    def set_planet_growth_specials(self):  # noqa complexity
        """Consider growth focus for planets with useful growth specials. Remove planets from list of candidates."""
        if not get_aistate().character.may_use_growth_focus():
            return

        # TODO Consider actual resource output of the candidate locations rather than only population
        for special, locations in get_growth_specials().items():
            # Find which metabolism is boosted by this special
            metabolism = AIDependencies.metabolismBoosts.get(special)
            if not metabolism:
                warning("Entry in available growth special not mapped to a metabolism")
                continue

            # Find the total population bonus we could get by using growth focus
            potential_pop_increase = empire_metabolisms.get(metabolism, 0)
            if not potential_pop_increase:
                continue

            debug(
                f"Considering setting growth focus for {special} at locations {locations} for potential population bonus of {potential_pop_increase:.1f}"
            )

            # Find the best suited planet to use growth special on, i.e. the planet where
            # we will lose the least amount of resource generation when using growth focus.
            def _print_evaluation(evaluation):
                """Local helper function printing a formatted evaluation."""
                debug(f"  - {planet} {evaluation}")

            ranked_planets = []
            for pid in locations:
                pinfo = self.planet_info.get(pid)
                planet = fo.getUniverse().getPlanet(pid)
                if not pinfo or GROWTH not in pinfo.planet.availableFoci:
                    _print_evaluation("has no growth focus available.")
                    continue

                # the increased population on the planet using this growth focus
                # is mostly wasted, so ignore it for now.
                pop = planet.currentMeterValue(fo.meterType.population)
                pop_gain = potential_pop_increase - planet.habitableSize
                if pop > pop_gain:
                    _print_evaluation(f"would lose more pop ({pop:.1f}) than gain everywhere else ({pop_gain:.1f}).")
                    continue

                # If we have a computronium special here, then research focus will have higher priority.
                if AIDependencies.COMPUTRONIUM_SPECIAL in planet.specials and RESEARCH in planet.availableFoci:
                    _print_evaluation("has a usable %s" % AIDependencies.COMPUTRONIUM_SPECIAL)
                    continue

                _print_evaluation(
                    f"considered (pop {pop:.1f}, growth gain {pop_gain:.1f}, current focus {pinfo.current_focus})"
                )

                # add a bias to discourage switching out growth focus to avoid focus change penalties
                if pinfo.current_focus == GROWTH:
                    pop -= 4

                ranked_planets.append((pop, pid, planet))

            if not ranked_planets:
                debug("  --> No suitable location found.")
                continue

            # sort possible locations by population in ascending order and set population
            # bonus at the planet with lowest possible population loss.
            ranked_planets.sort()
            for pop, pid, planet in ranked_planets:
                if self.bake_future_focus(pid, GROWTH):
                    debug("  --> Using growth focus at %s" % planet)
                    break
            else:
                warning("  --> Failed to set growth focus at all candidate locations.")

    def set_planet_production_and_research_specials(self):
        """Set production and research specials.
        Sets production/research specials for known (COMPUTRONIUM, HONEYCOMB and CONC_CAMP)
        production/research specials.
        Remove planets from list of candidates using bake_future_focus.
        """
        # loss should be >= 0 with 0 meaning research is the best focus anyway
        for loss, pid in sorted(
            [(self.planet_info[pid].rating_below_best(RESEARCH), pid) for pid in computronium_candidates()]
        ):
            # Computronium moon has a rather high stability requirement, in many cases it may not be worth anything
            if self.evaluate_computronium(pid, loss):
                self.bake_future_focus(pid, RESEARCH, True)
                break

        for loss, pid in sorted(
            [(self.planet_info[pid].rating_below_best(INDUSTRY), pid) for pid in honeycomb_candidates()]
        ):
            # honeycomb has no stability requirement, it should almost always be useful
            # TODO check for supply
            pp_gain = population_with_industry_focus() * get_named_real("HONEYCOMB_TARGET_INDUSTRY_PERPOP")
            if loss < pp_gain * self.priority_industry:
                self.bake_future_focus(pid, INDUSTRY, True)
                break
        # TODO concentration camps, currently the AI does not build them...

    def evaluate_computronium(self, pid: PlanetId, loss: float) -> bool:
        """
        Determine whether switching the given planet with a computronium to research focus is worth the loss.
        """
        researchers = 0.0  # amount of people that would profit from the special
        min_stability = get_named_real("COMPUTRONIUM_MIN_STABILITY")
        per_pop = get_named_real("COMPUTRONIUM_TARGET_RESEARCH_PERPOP")
        # TODO: check for supply as well
        for _, pinfo in self.planet_info.items():
            if pinfo.current_focus == RESEARCH and pinfo.possible_output[RESEARCH].stability >= min_stability:
                researchers += pinfo.planet.currentMeterValue(fo.meterType.population)
        return researchers * per_pop * self.priority_research > loss

    def set_influence_focus(self):
        """Assign planets to influence production."""
        current_ip = fo.getEmpire().resourceAvailable(fo.resourceType.influence)
        focused_planets = sorted(
            [
                (pinfo.evaluate(INFLUENCE), pinfo)
                for pinfo in self.planet_info.values()
                if pinfo.current_focus == INFLUENCE
            ]
        )
        candidate_planets = sorted(
            [
                (pinfo.evaluate(INFLUENCE), pinfo)
                for pinfo in self.planet_info.values()
                if pinfo.current_focus != INFLUENCE and INFLUENCE in pinfo.options
            ]
        )
        target_ip_production = sum(
            pi.possible_output[pi.current_focus].influence
            for pi in chain(self.planet_info.values(), self.baked_planet_info.values())
            if pi.current_focus in pi.possible_output
        )
        ratio = (current_ip + 10 * target_ip_production) / max(1.5, self.priority_influence)  # in turn 1 prio is 0
        debug(
            f"Evaluating influence: IP: {current_ip}, target production: {target_ip_production}, prio:"
            f"{self.priority_influence} => ratio: {ratio}\n"
            f"focused_planets: {focused_planets}\n"
            f"candidates: {candidate_planets}"
        )
        if candidate_planets and (ratio < 1.0 or ratio < 25.0 and candidate_planets[-1][0] > ratio):
            debug("setting influence focus to last candidate")
            self.bake_future_focus(candidate_planets[-1][1].planet.id, INFLUENCE)
        if focused_planets:
            pinfo = focused_planets[0][1]
            below_max = pinfo.possible_output[pinfo.current_focus].influence - pinfo.current_output.influence
            if below_max < 0.5 and (ratio > 30.0 or ratio > 2.0 and focused_planets[0][0] < -100.0 / ratio):
                debug("removing influence focus of first focus planet")
                pinfo = focused_planets.pop(0)
            # all (others) remain influence focussed.
            for _, pinfo in focused_planets:
                self.bake_future_focus(pinfo.planet.id, INFLUENCE)

        ProductionAI.candidate_for_translator = None
        translator_locations = BuildingType.TRANSLATOR.built_or_queued_at()
        for _, pinfo in focused_planets:
            if pinfo.planet.id not in translator_locations:
                ProductionAI.candidate_for_translator = pinfo.planet.id
                break

    def early_capital_handling(self):
        """
        A small, handcrafted start optimisation for bad researchers.
        Even if research is needed most, they do better by quickly finishing the Automatic History Analyzer.
        """
        if fo.currentTurn() < 7:
            capital = fo.getUniverse().getPlanet(fo.getEmpire().capitalID)
            if capital:
                factor = get_species_industry(capital.speciesName) / get_species_research(capital.speciesName)
                pinfo = self.planet_info[capital.id]
                if factor >= 2.0 and pinfo and pinfo.current_focus == INDUSTRY:
                    # factor at start is 2.66 for Egassem and 2 for good industry / bad research.
                    # 1.5 currently does not exist and for normal industry / bad research (1.33) the hack is no good.
                    switch_turn = 7 if factor > 2 else 6
                    if fo.currentTurn() < switch_turn:
                        debug("Special handling: keeping capital at industry.")
                        self.bake_future_focus(capital.id, INDUSTRY)
                    # else: don't do anything special, standard handling will likely switch to RESEARCH

    def immediate_loss_on_switch_to(self, pinfo: PlanetFocusInfo, focus: str) -> float:
        """
        When switching to focus, how much would the current production go down?
        E.g. when switching from research to production, how much does research go down?
        Since it may take several turns until we reach the new targets, this can be used as an indicator how
        'expensive' the switch would be.
        """
        # This is not totally correct, since it won't go down to the minimum immediately. but it should be a good
        # enough estimate to compare how expensive a different switches would be.
        low_rating = self.rate_output(
            pp=min(pinfo.current_output.industry, pinfo.possible_output[focus].industry),
            rp=min(pinfo.current_output.research, pinfo.possible_output[focus].research),
            ip=min(pinfo.current_output.influence, pinfo.possible_output[focus].influence),
            stability=min(pinfo.current_output.stability, pinfo.possible_output[focus].stability),
            focus=focus,
        )
        current_rating = self.rate_output(
            pp=pinfo.current_output.industry,
            rp=pinfo.current_output.research,
            ip=pinfo.current_output.influence,
            stability=pinfo.current_output.stability,
            focus=pinfo.current_focus,
        )

        # note that this may be zero, e.g. for a newly conquered planet or one that currently has negative stability
        return current_rating - low_rating

    def change_candidate(
        self, best_so_far: PlanetFocusInfo | None, candidate: PlanetFocusInfo, from_focus: str, to_focus: str
    ) -> PlanetFocusInfo:
        """
        Returns either candidate or best_so_far (if not None), depending on which one should rather do the
        given focus switch.
        """
        if best_so_far is None:
            return candidate
        best_val = best_so_far.possible_output[to_focus].rating - best_so_far.possible_output[from_focus].rating
        cand_val = candidate.possible_output[to_focus].rating - candidate.possible_output[from_focus].rating
        if best_val > 0:
            # to_focus would be an improvement for best_so_far, chose candidate if it gains even more
            return candidate if cand_val > best_val else best_so_far
        elif cand_val > 0:
            # to_focus would be an improvement for candidate
            return candidate
        else:
            # Neither would be better, prefer planets with a smaller immediate and long term loss
            # Small offsets, since both values may be zero
            best_val = (best_val - 0.1) * (self.immediate_loss_on_switch_to(best_so_far, to_focus) ** 0.5 + 0.1)
            cand_val = (cand_val - 0.1) * (self.immediate_loss_on_switch_to(candidate, to_focus) ** 0.5 + 0.1)
            return candidate if cand_val > best_val else best_so_far

    def get_planned_pp_rp_per_prio(self) -> (float, float, float, float):
        """
        Calculate ratios of planned PP and RP and both values divided by their priority.
        It uses future_focus of all planets, setting this value without baking it can be used to check alternatives.
        """
        planned_pp_target = sum(
            pi.possible_output[pi.future_focus].industry
            for pi in chain(self.baked_planet_info.values(), self.planet_info.values())
            if pi.future_focus in pi.possible_output
        )
        planned_rp_target = sum(
            pi.possible_output[pi.future_focus].research
            for pi in chain(self.baked_planet_info.values(), self.planet_info.values())
            if pi.future_focus in pi.possible_output
        )
        pp_per_priority = planned_pp_target / self.priority_industry
        # When the AI finished all research, research prio becomes 0
        rp_per_priority = planned_rp_target / self.priority_research if self.priority_research else 0.0
        debug(
            f"pp: {planned_pp_target}/{self.priority_industry}={pp_per_priority}, "
            f"rp: {planned_rp_target}/{self.priority_research}={rp_per_priority}"
        )
        return planned_pp_target, planned_rp_target, pp_per_priority, rp_per_priority

    def set_other_foci(self):
        """
        Set remaining planet's foci.
        If best focus is protection, set it, else if only one of industry and research is possible, set it.
        All others are set to industry or research, aiming for a production ratio like the ratio of the priorities.
        If we have one or more candidates currently set to neither industry nor research, we assign these, otherwise
        we may switch one planet from industry to research or vice versa, or let two planets swap their foci.
        """
        debug("set_other_foci...")
        may_switch_to_research = None
        may_switch_to_industry = None
        set_to_neither = []
        num_industry = num_research = 0
        for pid, pinfo in dict(self.planet_info).items():
            # Note that if focus hasn't been baked as INFLUENCE yet, it shouldn't be.
            if pinfo.rated_foci.best.focus in (INDUSTRY, RESEARCH, INFLUENCE) and {INDUSTRY, RESEARCH} <= pinfo.options:
                if pinfo.current_focus == INDUSTRY:
                    num_industry = num_industry + 1
                    may_switch_to_research = self.change_candidate(may_switch_to_research, pinfo, INDUSTRY, RESEARCH)
                elif pinfo.current_focus == RESEARCH:
                    num_research = num_research + 1
                    may_switch_to_industry = self.change_candidate(may_switch_to_industry, pinfo, RESEARCH, INDUSTRY)
                else:
                    set_to_neither.append(pinfo)
            else:
                # Either protection due to stability, or planet does not support both foci
                self.bake_future_focus(pid, pinfo.rated_foci.best.focus)
        if set_to_neither:
            self.chose_industry_or_research(set_to_neither)
        else:
            # Never switch all planets to the lower priority focus, especially not the capital while it is the only one
            if self.priority_industry > self.priority_research and num_industry < 2:
                may_switch_to_research = None
            if self.priority_research > self.priority_industry and num_research < 2:
                may_switch_to_industry = None
            self.chose_switches(may_switch_to_research, may_switch_to_industry)

    def chose_industry_or_research(self, set_to_neither: list[PlanetFocusInfo]) -> None:
        """Assign planets in set_to_neither to industry or research focus."""
        set_to_neither.sort()
        debug(f"chose_industry_or_research set_to_neither={set_to_neither}")
        while set_to_neither:
            # TBD: can we also use balanced_output_value here? Most likely its only one planet anyway...
            _, _, pp_per_priority, rp_per_priority = self.get_planned_pp_rp_per_prio()
            # sorting is best research to best industry
            if pp_per_priority > rp_per_priority:
                self.bake_future_focus(set_to_neither[0].planet.id, RESEARCH)
                set_to_neither.pop(0)
            else:
                self.bake_future_focus(set_to_neither[-1].planet.id, INDUSTRY)
                set_to_neither.pop()

    def chose_switches(self, to_research: PlanetFocusInfo | None, to_industry: PlanetFocusInfo | None) -> None:
        """
        Decide which of the two candidates should switch between industry and research, if any.
        """
        debug(f"chose_switches, candidates: research={to_research or '-'}, industry={to_industry or '-'}")
        tr_gain = to_research.difference(RESEARCH, INDUSTRY) if to_research else USELESS_RATING
        ti_gain = to_industry.difference(INDUSTRY, RESEARCH) if to_industry else USELESS_RATING
        if tr_gain > 0 and ti_gain > 0:
            debug("both are better with the new focus")
            self.bake_future_focus(to_research.planet.id, RESEARCH)
            self.bake_future_focus(to_industry.planet.id, INDUSTRY)
        else:
            pp, rp, pp_per_priority, rp_per_priority = self.get_planned_pp_rp_per_prio()
            current_best = self.balanced_output_value(self.priority_industry, pp, self.priority_research, rp)
            debug(f"current: {current_best:.2f}")
            if to_research:
                current_best = self.check_switch_for_ratio(to_research, RESEARCH, current_best)
            # TBD: can we skip the second call, if the first was an improvement?
            if to_industry:
                self.check_switch_for_ratio(to_industry, INDUSTRY, current_best)

    def check_switch_for_ratio(self, candidate: PlanetFocusInfo, new_focus: str, old_best: float) -> float:
        """
        Check whether a switch would improve the balanced output.
        If so, do the switch and return the new best value, else return old_best.
        """
        candidate.future_focus = new_focus
        pp, rp, _, _ = self.get_planned_pp_rp_per_prio()
        new_value = self.balanced_output_value(self.priority_industry, pp, self.priority_research, rp)
        debug(f"With switch to {new_focus}: {new_value:.2f}")
        if new_value < old_best:
            candidate.future_focus = candidate.current_focus
        self.bake_future_focus(candidate.planet.id, candidate.future_focus)
        return max(old_best, new_value)

    @staticmethod
    def balanced_output_value(prio1: float, production1: float, prio2: float, production2: float) -> float:
        """Evaluate possible production values with a preference for a production ratio equal to the priority ratio."""
        return prio1 * production1 + prio2 * production2 + min(prio1 * production2, prio2 * production1)


class Reporter:
    """Reporter contains some file scope functions to report"""

    def __init__(self, focus_manager):
        self.focus_manager = focus_manager
        self.sections = []
        self.captured_ids = set()

    def capture_section_info(self, title):
        """Grab ids of all the newly baked planets."""
        new_captured_ids = set(self.focus_manager.baked_planet_info)
        new_ids = new_captured_ids - self.captured_ids
        if new_ids:
            self.captured_ids = new_captured_ids
            self.sections.append((title, list(new_ids)))

    table_format = "%34s | %17s | %17s  | %13s | %13s  | %17s |"

    @staticmethod
    def print_resource_ai_header():
        debug("\n============================")
        debug("Collecting info to assess Planet Focus Changes\n")

    @staticmethod
    def print_table_header():
        debug("===================================")
        debug(
            Reporter.table_format,
            "Planet",
            "current RP/PP",
            "old target RP/PP",
            "current Focus",
            "newFocus",
            "new target RP/PP",
        )

    def print_table_footer(self, priority_ratio):
        current_industry_target = 0
        current_research_target = 0
        new_industry_target = 0
        new_research_target = 0
        all_industry_industry_target = 0
        all_industry_research_target = 0
        all_research_industry_target = 0
        all_research_research_target = 0
        total_changed = 0
        for pinfo in self.focus_manager.baked_planet_info.values():
            if pinfo.current_focus != pinfo.future_focus:
                total_changed += 1

            if pinfo.current_focus in pinfo.possible_output:
                old_pp, old_rp, *_ = pinfo.possible_output[pinfo.current_focus]
                current_industry_target += old_pp
                current_research_target += old_rp

            if pinfo.future_focus in pinfo.possible_output:
                future_pp, future_rp, *_ = pinfo.possible_output[pinfo.future_focus]
                new_industry_target += future_pp
                new_research_target += future_rp

            industry_pp, industry_rp, *_ = (
                pinfo.possible_output[INDUSTRY] if INDUSTRY in pinfo.possible_output else (future_pp, future_rp, 0, 0)
            )
            all_industry_industry_target += industry_pp
            all_industry_research_target += industry_rp

            research_pp, research_rp, *_ = (
                pinfo.possible_output[RESEARCH] if RESEARCH in pinfo.possible_output else (future_pp, future_rp, 0, 0)
            )
            all_research_industry_target += research_pp
            all_research_research_target += research_rp

        debug("-----------------------------------")
        debug(
            "Planet Focus Assignments to achieve target RP/PP ratio of %.2f"
            " from current target ratio of %.2f ( %.1f / %.1f )",
            priority_ratio,
            current_research_target / (current_industry_target + 0.0001),
            current_research_target,
            current_industry_target,
        )
        debug(
            "Max Industry assignments would result in target RP/PP ratio of %.2f ( %.1f / %.1f )",
            all_industry_research_target / (all_industry_industry_target + 0.0001),
            all_industry_research_target,
            all_industry_industry_target,
        )
        debug(
            "Max Research assignments would result in target RP/PP ratio of %.2f ( %.1f / %.1f )",
            all_research_research_target / (all_research_industry_target + 0.0001),
            all_research_research_target,
            all_research_industry_target,
        )
        debug("-----------------------------------")
        debug(
            "Final Ratio Target (turn %4d) RP/PP : %.2f ( %.1f / %.1f ) after %d Focus changes",
            fo.currentTurn(),
            new_research_target / (new_industry_target + 0.0001),
            new_research_target,
            new_industry_target,
            total_changed,
        )

    def print_table(self, priority_ratio):
        """Prints a table of all of the captured sections of assignments."""
        self.print_table_header()

        for title, id_set in self.sections:
            debug(
                Reporter.table_format,
                ("---------- " + title + " ------------------------------")[:33],
                "",
                "",
                "",
                "",
                "",
            )
            id_set.sort()  # pay sort cost only when printing
            for pid in id_set:
                pinfo = self.focus_manager.baked_planet_info[pid]
                old_focus = pinfo.current_focus
                new_focus = pinfo.future_focus
                current_pp, curren_rp, *_ = pinfo.current_output
                ot_pp, ot_rp, *_ = pinfo.possible_output.get(old_focus, (0, 0, 0, 0))
                nt_pp, nt_rp, *_ = pinfo.possible_output[new_focus]
                debug(
                    Reporter.table_format,
                    "pID (%3d) %22s" % (pid, pinfo.planet.name[-22:]),
                    f"c: {curren_rp:5.1f} / {current_pp:5.1f}",
                    f"cT: {ot_rp:5.1f} / {ot_pp:5.1f}",
                    "cF: %8s" % _focus_name(old_focus),
                    "nF: %8s" % _focus_name(new_focus),
                    f"cT: {nt_rp:5.1f} / {nt_pp:5.1f}",
                )
        self.print_table_footer(priority_ratio)

    @staticmethod
    def dump_output():
        empire = fo.getEmpire()
        pp, rp = empire.productionPoints, empire.resourceProduction(fo.resourceType.research)
        stats.output(fo.currentTurn(), rp, pp)

    @staticmethod
    def print_resources_priority():
        """Calculate top resource priority."""
        universe = fo.getUniverse()
        empire = fo.getEmpire()
        empire_planet_ids = PlanetUtilsAI.get_owned_planets_by_empire()
        debug("Resource Priorities:")
        resource_priorities = {}
        aistate = get_aistate()
        for priority_type in get_priority_resource_types():
            resource_priorities[priority_type] = aistate.get_priority(priority_type)

        sorted_priorities = sorted(resource_priorities.items(), key=itemgetter(1), reverse=True)
        top_priority = -1
        for evaluation_priority, evaluation_score in sorted_priorities:
            if top_priority < 0:
                top_priority = evaluation_priority
            debug("  %s: %.2f", evaluation_priority, evaluation_score)

        # what is the focus of available resource centers?
        debug("")
        warnings = {}
        foci_table = Table(
            Text("Planet"),
            Text("Size"),
            Text("Type"),
            Text("Focus"),
            Text("Species"),
            Text("Pop"),
            table_name="Planetary Foci Overview Turn %d" % fo.currentTurn(),
        )
        for pid in empire_planet_ids:
            planet = universe.getPlanet(pid)
            population = planet.currentMeterValue(fo.meterType.population)
            max_population = planet.currentMeterValue(fo.meterType.targetPopulation)
            if max_population < 1 and population > 0:
                warnings[planet.name] = (population, max_population)
            foci_table.add_row(
                planet,
                planet.size,
                planet.type,
                "_".join(str(planet.focus).split("_")[1:])[:8],
                planet.speciesName,
                f"{population:.1f}/{max_population:.1f}",
            )
        info(foci_table)
        debug(
            "Empire Totals:\nPopulation: %5d \nProduction: %5d\nResearch: %5d\n",
            empire.population(),
            empire.productionPoints,
            empire.resourceProduction(fo.resourceType.research),
        )
        for name, (cp, mp) in warnings.items():
            warning("Population Warning! -- %s has unsustainable current pop %d -- target %d", name, cp, mp)


def generate_resources_orders():
    """generate resources focus orders"""

    Reporter.print_resource_ai_header()
    resource_timer.start("Focus Infos")
    focus_manager = PlanetFocusManager()
    reporter = Reporter(focus_manager)
    focus_manager.set_planetary_foci(reporter)
    resource_timer.stop_print_and_clear()

    Reporter.dump_output()
    Reporter.print_resources_priority()



from typing import NamedTuple


class Bonus(NamedTuple):
    available: bool
    required_stability: float
    value: float

    def get_bonus(self, stability: float) -> float:
        if not self.available:
            return 0.0
        return self.value if stability >= self.required_stability else 0.0


def adjust_direction(threshold: float, current: float, target) -> int:
    """
    Returns -1, 0 or 1, depending on whether compared to current, target drops below, is on the same side or
    goes above the threshold.
    """
    return int(target >= threshold) - int(current >= threshold)



import pytest

from buildings import Shipyard


def test_get_shipyard_by_calling_constructor():
    assert Shipyard.GEO == Shipyard("BLD_SHIPYARD_CON_GEOINT")


def test_invalid_shipyard_raises_value_error_if_passed_via_constructor():
    with pytest.raises(ValueError, match="'XXX' is not a valid Shipyard"):
        Shipyard("XXX")



import copy
import freeOrionAIInterface as fo
from logging import debug, error, info, warning

from AIDependencies import INVALID_ID
from freeorion_tools import (
    assertion_fails,
    get_ship_part,
)
from freeorion_tools.design_compare import recursive_dict_diff
from turn_state import get_inhabited_planets


def build_cache_key(hullname: str, partlist: list[str]) -> str:
    """
    This reference name is used to identify existing designs and is mapped
    by Cache.map_reference_design_name to the ingame design name. Order of components are ignored.

    :param hullname: hull name
    :param partlist: list of part names
    :return: reference name
    """
    return "{}-{}".format(hullname, "-".join(sorted(partlist)))  # "Hull-Part1-Part2-Part3-Part4"


class ShipDesignCache:
    """This class handles the caching of information used to assess and build shipdesigns in this module.

    Important methods:
    update_for_new_turn(self): Updates the cache for the current turn, to be called once at the beginning of each turn.

    Important members:
    design_id_by_name          # {"designname": designid}
    part_by_partname           # {"partname": part object}
    map_reference_design_name  # {"reference_designname": "ingame_designname"}, cf. _build_reference_name()
    hulls_for_planets          # buildable hulls per planet {planetID: ["buildableHull1", "buildableHull2", ...]}
    parts_for_planets          # buildable parts per planet and slot: {planetID: {slottype1: ["part1", "part2"]}}
    best_designs               # {shipclass: {reqTup: {species: {available_parts: {hull: (rating, best_parts)}}}}}
    production_cost            # {planetID: {"partname1": local_production_cost, "hullname1": local_production_cost}}
    production_time            # {planetID: {"partname1": local_production_time, "hullname1": local_production_time}}

    Debug methods:
    print_CACHENAME(self), e.g. print_hulls_for_planets: prints content of the cache in some nicer format
    print_all(self): calls all the printing functions
    """

    def __init__(self):
        """Cache is empty on creation"""
        self.design_id_by_name = {}
        self.part_by_partname = {}
        self.map_reference_design_name = {}
        self.hulls_for_planets = {}
        self.parts_for_planets = {}
        self.best_designs = {}
        self.production_cost = {}
        self.production_time = {}
        self.last_printed = {}

    def update_for_new_turn(self):
        """Update the cache for the current turn.

        Make sure this function is called once at the beginning of the turn,
        i.e. before any other function of this module is used.
        """
        info(10 * "=" + "Updating ShipDesignCache for new turn" + 10 * "=")
        if not self.map_reference_design_name:
            self._build_cache_after_load()
        self._check_cache_for_consistency()
        self.update_cost_cache()
        self._update_buildable_items_this_turn()

    def print_design_id_by_name(self):
        """Print the design_id_by_name cache."""
        debug("DesignID cache: %s" % self.design_id_by_name)

    def print_part_by_partname(self):
        """Print the part_by_partname cache."""
        debug("Parts cached by name: %s" % self.part_by_partname)

    def print_map_reference_design_name(self):
        """Print the ingame, reference name map of shipdesigns."""
        debug("Design name map: %s" % self.map_reference_design_name)

    def print_hulls_for_planets(self, pid=None):
        """Print the hulls buildable on each planet.

        :param pid: None, int or list of ints
        """
        if pid is None:
            planets = list(self.hulls_for_planets)
        elif isinstance(pid, int):
            planets = [pid]
        elif isinstance(pid, list):
            planets = pid
        else:
            error("Invalid parameter 'pid' for 'print_hulls_for_planets'. Expected int, list or None.")
            return
        debug("Hull-cache:")
        get_planet = fo.getUniverse().getPlanet
        for pid in planets:
            debug(f"{get_planet(pid).name}: {self.hulls_for_planets[pid]}")

    def print_parts_for_planets(self, pid=None):
        """Print the parts buildable on each planet.

        :param pid: int or list of ints
        """
        if pid is None:
            planets = list(self.parts_for_planets)
        elif isinstance(pid, int):
            planets = [pid]
        elif isinstance(pid, list):
            planets = pid
        else:
            error("Invalid parameter 'pid' for 'print_parts_for_planets'. Expected int, list or None.")
            return
        debug("Available parts per planet:")
        get_planet = fo.getUniverse().getPlanet

        for pid in planets:
            debug("  %s:" % get_planet(pid).name)
            for slot in self.parts_for_planets[pid]:
                debug(f"    {slot}: {self.parts_for_planets[pid][slot]}")

    def print_best_designs(self, print_diff_only: bool = True):
        """Print the best designs that were previously found.

        :param print_diff_only: Print only changes to cache since last print
        """
        debug("Currently cached best designs:")
        if print_diff_only:
            print_dict = recursive_dict_diff(self.best_designs, self.last_printed, diff_level_threshold=1)
        else:
            print_dict = self.best_designs
        for classname in print_dict:
            debug(classname)
            cache_name = print_dict[classname]
            for consider_fleet in cache_name:
                debug(4 * " " + str(consider_fleet))
                cache_upkeep = cache_name[consider_fleet]
                for req_tuple in cache_upkeep:
                    debug(8 * " " + str(req_tuple))
                    cache_reqs = cache_upkeep[req_tuple]
                    for tech_tuple in cache_reqs:
                        debug(12 * " " + str(tech_tuple) + " # relevant tech upgrades")
                        cache_techs = cache_reqs[tech_tuple]
                        for species_tuple in cache_techs:
                            debug(16 * " " + str(species_tuple) + " # relevant species stats")
                            cache_species = cache_techs[species_tuple]
                            for av_parts in cache_species:
                                debug(20 * " " + str(av_parts))
                                cache_parts = cache_species[av_parts]
                                for hullname in sorted(cache_parts, reverse=True, key=lambda x: cache_parts[x][0]):
                                    debug(24 * " " + hullname + ":" + str(cache_parts[hullname]))
        self.last_printed = copy.deepcopy(self.best_designs)

    def print_production_cost(self):
        """Print production_cost cache."""
        universe = fo.getUniverse()
        debug("Cached production cost per planet:")
        for pid in self.production_cost:
            debug(f"  {universe.getPlanet(pid).name}: {self.production_cost[pid]}")

    def print_production_time(self):
        """Print production_time cache."""
        universe = fo.getUniverse()
        debug("Cached production cost per planet:")
        for pid in self.production_time:
            debug(f"  {universe.getPlanet(pid).name}: {self.production_time[pid]}")

    def print_all(self):
        """Print the entire ship design cache."""
        debug("Printing the ShipDesignAI cache...")
        self.print_design_id_by_name()
        self.print_part_by_partname()
        self.print_map_reference_design_name()
        self.print_hulls_for_planets()
        self.print_parts_for_planets()
        self.print_best_designs()
        self.print_production_cost()
        self.print_production_time()

    def update_cost_cache(self, partnames=None, hullnames=None):
        """Cache the production cost and time for each part and hull for each inhabited planet for this turn.

        If partnames and hullnames are both None, rebuild Cache with available parts.
        Otherwise, update cache for the specified items.

        :param partnames: iterable
        :param hullnames: iterable
        """
        empire = fo.getEmpire()
        empire_id = empire.empireID

        parts_to_update = set()
        hulls_to_update = set()
        if partnames is None and hullnames is None:
            # clear cache and rebuild with available parts, called at the beginning of the turn
            self.production_cost.clear()
            self.production_time.clear()
            parts_to_update.update(list(empire.availableShipParts))
            hulls_to_update.update(list(empire.availableShipHulls))
        if partnames:
            parts_to_update.update(partnames)
        if hullnames:
            hulls_to_update.update(hullnames)

        # no need to update items we already cached in this turn
        pids = list(get_inhabited_planets())
        if self.production_cost and pids:
            cached_items = set(self.production_cost[pids[0]].keys())
            parts_to_update -= cached_items
            hulls_to_update -= cached_items

        for partname in parts_to_update:
            part = get_ship_part(partname)
            for pid in pids:
                self.production_cost.setdefault(pid, {})[partname] = part.productionCost(empire_id, pid, INVALID_ID)
                self.production_time.setdefault(pid, {})[partname] = part.productionTime(empire_id, pid, INVALID_ID)
        for hullname in hulls_to_update:
            hull = fo.getShipHull(hullname)
            for pid in pids:
                self.production_cost.setdefault(pid, {})[hullname] = hull.productionCost(empire_id, pid, INVALID_ID)
                self.production_time.setdefault(pid, {})[hullname] = hull.productionTime(empire_id, pid, INVALID_ID)

    def _build_cache_after_load(self):
        """Build cache after loading or starting a game.

        This function is supposed to be called after a reload or at the first turn.
        It reads out all the existing ship designs and then updates the following cache:
        - map_reference_design_name
        - design_id_by_name
        """
        if self.map_reference_design_name or self.design_id_by_name:
            warning("ShipDesignAI.Cache._build_cache_after_load() called but cache is not empty.")
        for design_id in fo.getEmpire().allShipDesigns:
            design = fo.getShipDesign(design_id)
            reference_name = build_cache_key(design.hull, design.parts)
            self.map_reference_design_name[reference_name] = design.name
            self.design_id_by_name[design.name] = design_id

    def _check_cache_for_consistency(self):  # noqa: C901
        """Check if the persistent cache is consistent with the gamestate and fix it if not.

        This function should be called once at the beginning of the turn (before update_shipdesign_cache()).
        Especially (only?) in multiplayer games, the shipDesignIDs may sometimes change across turns.
        """
        debug("Checking persistent cache for consistency...")
        try:
            for partname in self.part_by_partname:
                cached_name = self.part_by_partname[partname].name
                if cached_name != partname:
                    self.part_by_partname[partname] = fo.getShipPart(partname)
                    error(f"Part cache corrupted. Expected: {partname}, got: {cached_name}. Cache was repaired.")
        except Exception as e:
            self.part_by_partname.clear()
            error(e, exc_info=True)

        corrupted = []
        # create a copy of the dict-keys so we can alter the dict
        for designname in list(self.design_id_by_name):
            # dropping invalid designs from cache
            if self.design_id_by_name[designname] == INVALID_ID:
                del self.design_id_by_name[designname]
                continue
            try:
                cached_name = fo.getShipDesign(self.design_id_by_name[designname]).name
                if cached_name != designname:
                    warning(f"ShipID cache corrupted. Expected: {designname}, got: {cached_name}.")
                    design_id = next(
                        iter(
                            [
                                shipDesignID
                                for shipDesignID in fo.getEmpire().allShipDesigns
                                if designname == fo.getShipDesign(shipDesignID).name
                            ]
                        ),
                        None,
                    )
                    if design_id is not None:
                        self.design_id_by_name[designname] = design_id
                    else:
                        corrupted.append(designname)
            except AttributeError:
                warning("ShipID cache corrupted. Could not get cached shipdesign. Repairing Cache.", exc_info=True)
                design_id = next(
                    iter(
                        [
                            shipDesignID
                            for shipDesignID in fo.getEmpire().allShipDesigns
                            if designname == fo.getShipDesign(shipDesignID).name
                        ]
                    ),
                    None,
                )
                if design_id is not None:
                    self.design_id_by_name[designname] = design_id
                else:
                    corrupted.append(designname)
        for corrupted_entry in corrupted:
            del self.design_id_by_name[corrupted_entry]
            bad_ref = next(
                iter([_key for _key, _val in self.map_reference_design_name.items() if _val == corrupted_entry]), None
            )
            if bad_ref is not None:
                del self.map_reference_design_name[bad_ref]

    def _update_buildable_items_this_turn(self):
        """Calculate which parts and hulls can be built on each planet this turn."""
        self.hulls_for_planets.clear()
        self.parts_for_planets.clear()
        empire = fo.getEmpire()
        all_hulls = list(empire.availableShipHulls)
        all_parts = list(empire.availableShipParts)

        for pid in get_inhabited_planets():
            for hull_name in all_hulls:
                hull = fo.getShipHull(hull_name)
                if assertion_fails(hull is not None):
                    continue

                if hull.productionLocation(pid):
                    self.hulls_for_planets.setdefault(pid, []).append(hull_name)

            for part_name in all_parts:
                ship_part = get_ship_part(part_name)
                if assertion_fails(ship_part is not None):
                    continue

                slot_types = ship_part.mountableSlotTypes
                if ship_part.productionLocation(pid):
                    for slot_type in slot_types:
                        self.parts_for_planets.setdefault(pid, {}).setdefault(slot_type, []).append(part_name)



import freeOrionAIInterface as fo

import AIDependencies
from AIDependencies import (
    POP_CONST_MOD_MAP,
    POP_SIZE_MOD_MAP_MODIFIED_BY_SPECIES,
    POP_SIZE_MOD_MAP_NOT_MODIFIED_BY_SPECIES,
)
from colonization.colony_score import debug_rating
from freeorion_tools import get_species_population, tech_is_complete

active_growth_specials = {}


def calc_max_pop(planet, species, detail):  # noqa: C901
    planet_size = planet.habitableSize
    planet_env = species.getPlanetEnvironment(planet.type)
    if planet_env == fo.planetEnvironment.uninhabitable:
        detail.append("Uninhabitable.")
        return 0

    for bldg_id in planet.buildingIDs:
        building = fo.getUniverse().getBuilding(bldg_id)
        if not building:
            continue
        if building.buildingTypeName == "BLD_GATEWAY_VOID":
            detail.append("Gateway to the void: Uninhabitable.")
            return 0

    if planet.speciesName in AIDependencies.SPECIES_FIXED_POPULATION:
        return AIDependencies.SPECIES_FIXED_POPULATION[planet.speciesName]

    tag_list = list(species.tags) if species else []
    pop_tag_mod = get_species_population(species.name)
    if planet.type == fo.planetType.gasGiant and "GASEOUS" in tag_list:
        gaseous_adjustment = AIDependencies.GASEOUS_POP_FACTOR
        detail.append("GASEOUS adjustment: %.2f" % gaseous_adjustment)
    else:
        gaseous_adjustment = 1.0

    planet_specials = set(planet.specials)

    base_pop_modified_by_species = 0
    base_pop_not_modified_by_species = 0
    pop_const_mod = 0

    # first, account for the environment
    environment_mod = POP_SIZE_MOD_MAP_MODIFIED_BY_SPECIES["environment_bonus"][planet_env]
    detail.append("Base environment: %d" % environment_mod)
    base_pop_modified_by_species += environment_mod

    # find all applicable modifiers
    for tech in POP_SIZE_MOD_MAP_MODIFIED_BY_SPECIES:
        if tech != "environment_bonus" and tech_is_complete(tech):
            base_pop_modified_by_species += POP_SIZE_MOD_MAP_MODIFIED_BY_SPECIES[tech][planet_env]
            detail.append("%s_PSM_early(%d)" % (tech, POP_SIZE_MOD_MAP_MODIFIED_BY_SPECIES[tech][planet_env]))

    for tech in POP_SIZE_MOD_MAP_NOT_MODIFIED_BY_SPECIES:
        if tech_is_complete(tech):
            base_pop_not_modified_by_species += POP_SIZE_MOD_MAP_NOT_MODIFIED_BY_SPECIES[tech][planet_env]
            detail.append("%s_PSM_late(%d)" % (tech, POP_SIZE_MOD_MAP_NOT_MODIFIED_BY_SPECIES[tech][planet_env]))

    for tech in POP_CONST_MOD_MAP:
        if tech_is_complete(tech):
            pop_const_mod += POP_CONST_MOD_MAP[tech][planet_env]
            detail.append("%s_PCM(%d)" % (tech, POP_CONST_MOD_MAP[tech][planet_env]))

    for _special in planet_specials.intersection(AIDependencies.POP_FIXED_MOD_SPECIALS):
        if AIDependencies.not_affect_by_special(_special, species.name):
            continue
        this_mod = AIDependencies.POP_FIXED_MOD_SPECIALS[_special]
        detail.append("%s_PCM(%d)" % (_special, this_mod))
        pop_const_mod += this_mod

    for _special in planet_specials.intersection(AIDependencies.POP_PROPORTIONAL_MOD_SPECIALS):
        this_mod = AIDependencies.POP_PROPORTIONAL_MOD_SPECIALS[_special]
        detail.append(f"{_special} (maxPop{this_mod:+.1f})")
        base_pop_not_modified_by_species += this_mod

    gaia = AIDependencies.GAIA_SPECIAL
    if gaia in planet_specials and not AIDependencies.not_affect_by_special(gaia, species):
        base_pop_not_modified_by_species += 3
        detail.append("Gaia_PSM_late(3)")

    if "SELF_SUSTAINING" in tag_list:
        if planet_env == fo.planetEnvironment.good:
            base_pop_not_modified_by_species += 3
            detail.append("SelfSustaining_PSM_late(3)")

    applicable_boosts = set()
    for this_tag in [tag for tag in tag_list if tag in AIDependencies.metabolismBoostMap]:
        metab_boosts = AIDependencies.metabolismBoostMap.get(this_tag, [])
        for key in active_growth_specials.keys():
            if len(active_growth_specials[key]) > 0 and key in metab_boosts:
                applicable_boosts.add(key)
                detail.append("%s boost active" % key)
        for boost in metab_boosts:
            if boost in planet_specials:
                applicable_boosts.add(boost)
                detail.append("%s boost present" % boost)

    n_boosts = len(applicable_boosts)
    if n_boosts:
        base_pop_not_modified_by_species += n_boosts
        detail.append("boosts_PSM(%d from %s)" % (n_boosts, applicable_boosts))

    if planet.id in species.homeworlds:
        detail.append("Homeworld (2)")
        base_pop_not_modified_by_species += 2

    if AIDependencies.TAG_LIGHT_SENSITIVE in tag_list:
        star_type = fo.getUniverse().getSystem(planet.systemID).starType
        star_pop_mod = AIDependencies.POP_MOD_LIGHTSENSITIVE_STAR_MAP.get(star_type, 0)
        base_pop_not_modified_by_species += star_pop_mod
        detail.append("Lightsensitive Star Bonus_PSM_late(%.1f)" % star_pop_mod)

    def max_pop_size():
        species_effect = (pop_tag_mod - 1) * abs(base_pop_modified_by_species)
        gaseous_effect = (gaseous_adjustment - 1) * abs(base_pop_modified_by_species)
        base_pop = base_pop_not_modified_by_species + base_pop_modified_by_species + species_effect + gaseous_effect
        return planet_size * base_pop + pop_const_mod

    target_pop = max_pop_size()
    if "PHOTOTROPHIC" in tag_list and target_pop > 0:
        star_type = fo.getUniverse().getSystem(planet.systemID).starType
        star_pop_mod = AIDependencies.POP_MOD_PHOTOTROPHIC_STAR_MAP.get(star_type, 0)
        base_pop_not_modified_by_species += star_pop_mod
        detail.append("Phototropic Star Bonus_PSM_late(%0.1f)" % star_pop_mod)
        target_pop = max_pop_size()

    detail.append("max_pop = base + size*[psm_early + species_mod*abs(psm_early) + psm_late]")
    detail.append(
        "        = %.2f + %d * [%.2f + %.2f*abs(%.2f) + %.2f]"
        % (
            pop_const_mod,
            planet_size,
            base_pop_modified_by_species,
            (pop_tag_mod + gaseous_adjustment - 2),
            base_pop_modified_by_species,
            base_pop_not_modified_by_species,
        )
    )
    detail.append("        = %.2f" % target_pop)
    # new rating code passes a dummy to this function and only adds the result to detail
    debug_rating(f"Details from calc_max_pop: {detail}")
    return target_pop



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="PRO_EXOBOTS",
    description="PRO_EXOBOTS_DESC",
    short_description="EXOBOT_SHORT_DESC",
    category="PRODUCTION_CATEGORY",
    researchcost=96 * TECH_COST_MULTIPLIER,
    researchturns=6,
    tags=["PEDIA_PRODUCTION_CATEGORY"],
    prerequisites=["LRN_NASCENT_AI", "PRO_NANOTECH_PROD"],
    unlock=[Item(type=UnlockBuilding, name="BLD_COL_EXOBOT"), Item(type=UnlockPolicy, name="PLC_EXOBOT_PRODUCTIVITY")],
    graphic="icons/species/robotic-01.png",
)



import freeorion as fo
import sys
from collections import defaultdict
from math import acos, ceil, cos, floor, pi, sin, sqrt
from random import gauss, randint, random, uniform

import universe_tables
import util


class AdjacencyGrid:
    def __init__(self, universe_width):
        """
        :param float universe_width:
        """
        assert universe_tables.MIN_SYSTEM_SEPARATION < universe_tables.MAX_STARLANE_LENGTH
        self.min_dist = universe_tables.MIN_SYSTEM_SEPARATION
        self.max_dist = universe_tables.MAX_STARLANE_LENGTH
        self.cell_size = min(max(universe_width / 50, self.min_dist), self.max_dist / sqrt(2))
        self.width = int(universe_width / self.cell_size) + 1
        self.grid = defaultdict(set)
        print(f"Adjacency Grid: width {self.width}, cell size {self.cell_size}")

    def cell(self, pos):
        """Returns cell index."""
        return int(pos[0] / self.cell_size), int(pos[1] / self.cell_size)

    def insert_pos(self, pos):
        """Inserts pos."""
        self.grid[self.cell(pos)].add(pos)

    def remove_pos(self, pos):
        """Removes pos."""
        self.grid[self.cell(pos)].discard(pos)

    def _square_indices_containing_cell(self, cell, radius):
        """Return a square of indices containing cell."""
        cell_x, cell_y = cell
        upper_left_x = max(0, cell_x - radius)
        upper_left_y = max(0, cell_y - radius)
        lower_right_x = min(self.width - 1, cell_x + radius)
        lower_right_y = min(self.width - 1, cell_y + radius)
        return [(x, y) for x in range(upper_left_x, lower_right_x + 1) for y in range(upper_left_y, lower_right_y + 1)]

    def _generate_rings_around_point(self, pos, max_radius=None):
        """Yields successive rings of indices containing cell."""
        cell = self.cell(pos)
        i_radius = 0
        i_radius_max = ceil(max_radius / self.cell_size) if max_radius else (self.width + 1)
        outer = set()
        ring = {1}
        while ring and i_radius <= i_radius_max:
            inner = outer
            outer = set(self._square_indices_containing_cell(cell, i_radius))
            ring = outer - inner
            yield ring
            i_radius += 1

    def neighbors(self, p):
        """Return all neighbors within max_dist."""
        _neighbors = []
        i_radius_close_enough = floor(self.max_dist / self.cell_size)
        for i, ring in enumerate(self._generate_rings_around_point(p, self.max_dist)):
            if i <= i_radius_close_enough:
                _neighbors += [pos for cell in ring for pos in self.grid[cell]]
            else:
                _neighbors += [
                    pos for cell in ring for pos in self.grid[cell] if util.distance(pos, p) <= self.max_dist
                ]
        return _neighbors

    def nearest_neighbor(self, p):
        """Return nearest neighbor at any distance."""
        for ring in self._generate_rings_around_point(p):
            candidates = [pos for cell in ring for pos in self.grid[cell]]
            if candidates:
                (_, pt) = min((util.distance(pos, p), pos) for pos in candidates)
                return pt
        return None

    def too_close_to_other_positions(self, pos):
        """
        Checks if the specified position is too close to the positions stored in the grid
        """
        pos_cell = self.cell(pos)
        return any(
            util.distance(p2, pos) < self.min_dist
            for p2_cell in self._square_indices_containing_cell(pos_cell, 1)
            for p2 in self.grid[p2_cell]
        )


class DSet:
    """
    A set of pos that is disjoint (not connected) to other sets.

    If the set is of one item then it has no parent. parent equals None.
    Otherwise the set points to the root/representative item of the set via parent.

    A private child class of DisjointSets to store the parent and rank of a disjoint set object
    """

    def __init__(self, pos):
        self.pos = pos

        # stores None or the parent
        self.parent = None

        # starts at 0, never decreases and only increases
        # when this object is the parent and two equal rank sets are merged
        self.rank = 0

    def bind_parent(self, parent):
        """Bind to parent."""
        assert self is not parent
        self.parent = parent

    def inc_rank(self):
        """Increment rank."""
        self.rank += 1


class DisjointSets:
    """
    A set of disjoint sets.

    It supports the operations of:
    add(pos)      -- Creates a new single item set containing pos if it isn't already a set. O(1)
    link(p1, p2)  -- Links two sets containing p1 and p2 together.
                     This will add(p1) and add(p2) if necessary. O(1)
    root(pos)     -- Gets the root/representative object for the set.
                     Creates a new set with pos if not already a set. O(1)
    complete_sets -- Returns a list of list of all sets. O(n)

    See https://en.wikipedia.org/wiki/Disjoint-set_data_structure
    """

    def __init__(self):
        self.dsets = {}

    def add(self, pos):
        """Creates a new single item set containing pos if it isn't already a set. O(1)"""
        if pos not in self.dsets:
            self.dsets[pos] = DSet(pos)

    def link(self, p1, p2):
        """Creates a link between the sets containing p1 and p2"""
        root1 = self.root(p1)
        if p1 == p2:
            # print " self link"
            return
        root2 = self.root(p2)
        if root1 == root2:
            # print " same discrete set link"
            return
        ds1 = self.dsets[root1]
        ds2 = self.dsets[root2]
        assert not ds1.parent and not ds2.parent

        # Always attach the "smaller" lower rank as the child of the "larger" higher rank tree
        if ds1.rank > ds2.rank:
            # print " left link"
            ds2.bind_parent(ds1)
        else:
            # print " right link"
            ds1.bind_parent(ds2)
            if ds1.rank == ds2.rank:
                # print " even link"
                ds2.inc_rank()

    def root(self, pos):
        """
        Returns the key position of the cluster containing pos
        Adds pos if absent
        """
        root = self._has_root(pos)
        if root:
            return root[0]
        self.add(pos)
        return pos

    def _has_root(self, pos):
        """
        Check if pos is the root of a cluster otherwise shorten
        the tree while tranversing it and return the root
        """

        # traverse tree and fetch parents for compression
        def parents(p1, children):
            # print "pp p1 ", p1.pos, " children a ", children
            if p1.parent:
                # print "pp deeper ", p1.pos, " chlidren b ", children
                children.append(p1)
                return parents(p1.parent, children)
            else:
                # print "pp done p1 ", p1.pos, " children c ", children
                return (p1, children)

        if pos not in self.dsets:
            return []

        (root, children) = parents(self.dsets[pos], [])
        # squash the chain of children
        for child in children:
            child.bind_parent(root)
        return [root.pos]

    def complete_sets(self):
        """returns a list of list of all sets O(n)."""
        ret = defaultdict(list)
        for pos in self.dsets.keys():
            ret[self.root(pos)].append(pos)
        return list(ret.values())


class Clusterer:
    """
    Computes all clusters of positions separated by more than MAX_STARLANE_LENGTH

    This creates a disjoint set of all of the positions
    Adds all of the seperations less than MAX_STARLANE_LENGTH
    Then returns the clusters of stars when requested
    """

    def __init__(self, positions, adjacency_grid):
        self.adjacency_grid = adjacency_grid
        self.dsets = DisjointSets()

        for pos in positions:
            self._add(pos)

        self.clusters = set()
        for cluster in self.dsets.complete_sets():
            # Decorate each cluster with its length to speed sorting
            self.clusters.add((len(cluster), frozenset(cluster)))

    def __len__(self):
        return len(self.clusters)

    def _add(self, pos):
        """Adds pos."""
        for neighbor in self.adjacency_grid.neighbors(pos):
            # print "Clusterer adding ", pos,  " and ", neighbor
            self.dsets.link(pos, neighbor)

    def smallest_isolated_cluster(self):
        """
        Return None if all of the positions are in one cluster
        or the smallest cluster
        """
        # Sort (min) and return the undecorated cluster.
        return None if len(self.clusters) < 2 else min(self.clusters)[1]

    def stitch_clusters(self, p1, p2, stitches):
        """
        Stitch the clusters containing ''p1'' and ''p2''
        together with the positions in ''stitches''

        This assumes that the stitching is correctly formed.
        ''p1'' and ''p2'' are the closest positions between
        the clusters and the positions in ''stitches'' are only
        between ''p1'' and ''p2''

        After stitch_clusters there will be one fewer clusters in
        the clusterer, unless there were fewer than two clusters
        to begin with.

        If ''p1'' or ''p2'' are not in clusters then stitch_clusters
        just stitches two random clusters together.  This preserves
        the invariant that stitch_clusters always reduces the number
        of clusters.
        """
        len_dset1 = None
        len_dset2 = None
        for len_dset in self.clusters:
            if p1 in len_dset[1]:
                len_dset1 = len_dset
            if p2 in len_dset[1]:
                len_dset2 = len_dset

        if not len_dset1 or not len_dset2:
            util.report_error("p1 and p2 must be points in disjoint sets of positions")
            if len(self.clusters) < 2:
                return
            len_dset1, len_dset2 = list(self.clusters)[0:2]

        self.clusters.remove(len_dset1)
        self.clusters.remove(len_dset2)
        # Decorate the new cluster with its length to speed sorting
        new_set = len_dset1[1].union(len_dset2[1]).union(frozenset(stitches))
        self.clusters.add((len(new_set), new_set))


def stitching_positions(p1, p2):
    """
    Returns a list of positions between p1 and p2 between MIN_SYSTEM_SEPARATION and MAX_STARLANE_LENGTH apart
    """
    if 2 * universe_tables.MIN_SYSTEM_SEPARATION >= universe_tables.MAX_STARLANE_LENGTH:
        util.report_error(
            "MAX_STARLANE_LENGTH must be twice MIN_SYSTEM_SEPARATION to "
            "allow extra positions to be added to enforce MAX_STARLANE_LENGTH"
        )
        return []

    max_dist = universe_tables.MAX_STARLANE_LENGTH
    min_dist = universe_tables.MIN_SYSTEM_SEPARATION
    p1_p2_dist = util.distance(p1, p2)
    if p1_p2_dist < max_dist:
        return []

    # Pick a random point in an 2:1 ratio ellipse and then rotate and slide it in between p1 and p2
    p1_p2_theta = acos((p2[0] - p1[0]) / p1_p2_dist)
    p1_p2_scale = (p1_p2_dist - 2 * min_dist) / 2
    p1_p2_translation = ((p1[0] + p2[0]) / 2, (p1[1] + p2[1]) / 2)

    radius_0space = uniform(0.0, 1.0)
    angle_0space = uniform(0.0, 2.0 * pi)
    rotated_point = (
        radius_0space * p1_p2_scale * cos(angle_0space + p1_p2_theta),
        radius_0space * p1_p2_scale * sin(angle_0space + p1_p2_theta),
    )

    p3 = (rotated_point[0] + p1_p2_translation[0], rotated_point[1] + p1_p2_translation[1])

    return [p3] + stitching_positions(p1, p3) + stitching_positions(p2, p3)


def enforce_max_distance(positions, adjacency_grid):
    """
    Adds extra positions between groups of positions to guarantee
    that no groups of positions are separated from all other groups
    of positions by more than universe_tables.MAX_STARLANE_LENGTH

    Find all groups of positions, where every member is within
    MAX_STARLANE_LENGTH of at least one other member of the group,
    but all members of the group are more than MAX_STARLANE_LENGTH
    from all positions in other groups.

    For each group:
      - Remove it from the adjacency grid and find its nearest neighboring position.
      - Add positions between the nearest neighbor and the group to
        ensure every position has at least one neighbor within
        MAX_STARLANE_LENGTH.
    """
    # Find all clusters
    clusterer = Clusterer(positions, adjacency_grid)

    if len(clusterer) == 1:
        print("All systems positioned in a single connected cluster")
    else:
        print(f"{len(clusterer)} clusters separated by more than the MAX_STARLANE_LENGTH." "  Starting to fill gaps.")

    while len(clusterer) > 1:
        smallest_cluster = clusterer.smallest_isolated_cluster()
        print("Searching for nearest neighbor position to a cluster " f"with {len(smallest_cluster)} positions.")
        for pos in smallest_cluster:
            adjacency_grid.remove_pos(pos)
        # Find nearest neighbour
        (_, p1, p2) = min(
            [
                (util.distance(pos, nn), pos, nn)
                for pos in smallest_cluster
                for nn in [adjacency_grid.nearest_neighbor(pos)]
                if nn
            ]
        )

        for pos in smallest_cluster:
            adjacency_grid.insert_pos(pos)
        # Add extra positions
        extra_positions = stitching_positions(p1, p2)
        for i_extra, p3 in enumerate(extra_positions):
            print(
                f"Adding {i_extra + 1} of {len(extra_positions)} extra positions at {p3} to enforce "
                f"max spacing between {p1} and {p2}."
            )
            adjacency_grid.insert_pos(p3)
            positions.append(p3)
        clusterer.stitch_clusters(p1, p2, extra_positions)

        if len(clusterer) == 1:
            print("All systems now positioned in a single connected cluster")
        else:
            print(f"{len(clusterer)} clusters separated by more the MAX_STARLANE_LENGTH.  " "Continuing to fill gaps.")


def calc_universe_width(shape, size):
    """
    Calculates typical universe width based on galaxy shape and number of star systems.
    A 150 star universe should be 1000 units across.
    """
    width = (1000.0 / sqrt(150.0)) * sqrt(float(size))
    if shape in [fo.galaxyShape.elliptical, fo.galaxyShape.irregular]:
        width *= 1.4
    elif shape == fo.galaxyShape.disc:
        width *= 1.2
    return width


def spiral_galaxy_calc_positions(positions, adjacency_grid, arms, size, width):
    """
    Calculate positions for the spiral galaxy shapes.
    """
    arm_offset = uniform(0.0, 2.0 * pi)
    arm_angle = (2.0 * pi) / float(arms)
    arm_spread = (0.3 * pi) / float(arms)
    arm_length = 1.5 * pi
    center = 0.25

    for i in range(size):
        attempts = 100
        while attempts > 0:
            radius = random()
            if radius < center:
                angle = uniform(0.0, 2.0 * pi)
                x = radius * cos(arm_offset + angle)
                y = radius * sin(arm_offset + angle)
            else:
                arm = randint(0, arms - 1) * arm_angle
                angle = gauss(0.0, arm_spread)
                x = radius * cos(arm_offset + arm + angle + radius * arm_length)
                y = radius * sin(arm_offset + arm + angle + radius * arm_length)

            x = (x + 1) * width / 2.0
            y = (y + 1) * width / 2.0
            if (x < 0.0) or (width <= x) or (y < 0.0) or (width <= y):
                continue

            # see if new star is too close to any existing star.
            # if so, we try again or give up
            if adjacency_grid.too_close_to_other_positions((x, y)):
                attempts -= 1
                continue

            # add the new star location
            pos = (x, y)
            adjacency_grid.insert_pos(pos)
            positions.append(pos)
            break

        if not attempts:
            print(
                f"Spiral galaxy shape: giving up on placing star {i}, can't "
                "find position sufficiently far from other systems."
            )


def elliptical_galaxy_calc_positions(positions, adjacency_grid, size, width):
    """
    Calculate positions for the elliptical galaxy shape.
    """
    ellipse_width_vs_height = uniform(0.4, 0.6)
    rotation = uniform(0.0, pi)
    rotation_sin = sin(rotation)
    rotation_cos = cos(rotation)
    gap_constant = 0.95
    gap_size = 1.0 - gap_constant * gap_constant * gap_constant

    for i in range(size):
        attempts = 100
        while attempts > 0:
            radius = uniform(0.0, gap_constant)
            # adjust for bigger density near center and create gap
            radius = radius * radius * radius + gap_size
            angle = uniform(0.0, 2.0 * pi)

            # rotate for individual angle and apply elliptical shape
            x1 = radius * cos(angle)
            y1 = radius * sin(angle) * ellipse_width_vs_height

            # rotate for ellipse angle
            x = x1 * rotation_cos - y1 * rotation_sin
            y = x1 * rotation_sin + y1 * rotation_cos

            # move from [-1.0, 1.0] universe coordinates
            x = (x + 1.0) * width / 2.0
            y = (y + 1.0) * width / 2.0

            # discard stars that are outside boundaries (due to possible rounding errors)
            if (x < 0) or (x >= width) or (y < 0) or (y >= width):
                attempts -= 1
                continue

            # see if new star is too close to any existing star; if so, we try again
            if adjacency_grid.too_close_to_other_positions((x, y)):
                attempts -= 1
                continue

            # add the new star location
            pos = (x, y)
            adjacency_grid.insert_pos(pos)
            positions.append(pos)
            break

        if not attempts:
            print(
                f"Elliptical galaxy shape: giving up on placing star {i}, "
                "can't find position sufficiently far from other systems"
            )


def disc_galaxy_calc_positions(positions, adjacency_grid, size, width):
    """
    Calculate positions for the disc galaxy shape.
    """
    center_x, center_y = width / 2.0, width / 2.0

    for i in range(size):
        attempts = 100
        while attempts > 0:
            radius = uniform(0.0, width / 2.0)
            angle = uniform(0.0, 2.0 * pi)
            x = center_x + radius * cos(angle)
            y = center_y + radius * sin(angle)

            if (x < 0) or (width <= x) or (y < 0) or (width <= y):
                attempts -= 1
                continue

            # see if new star is too close to any existing star; if so, we try again
            if adjacency_grid.too_close_to_other_positions((x, y)):
                attempts -= 1
                continue

            # add the new star location
            pos = (x, y)
            adjacency_grid.insert_pos(pos)
            positions.append(pos)
            break

        if not attempts:
            print(
                f"Disc galaxy shape: giving up on placing star {i}, can't find"
                " position sufficiently far from other systems"
            )


def cluster_galaxy_calc_positions(  # noqa: C901
    positions: list[tuple[int, int]], adjacency_grid: AdjacencyGrid, size: int, width: float
):
    """
    Calculate positions for the cluster galaxy shape and add them to positions.
    """
    if size < 1:
        print("Cluster galaxy requested for less than 1 star", file=sys.stderr)
        return

    if size == 1:
        pos = (width / 2.0, width / 2.0)
        positions.append(pos)
        return

    # Typically a galaxy with 100 systems should have ~5 clusters
    avg_clusters = max(2, round(size / 20))
    # Add a bit of random variation (+/- 20%)
    clusters = max(2, randint(round((avg_clusters * 8) / 10), round((avg_clusters * 12) / 10)))

    # probability of systems which don't belong to a cluster
    system_noise = 0.15
    ellipse_width_vs_height = uniform(0.2, 0.5)
    # a list of tuples of tuples: first tuple holds cluster position,
    # second tuple stores help values for cluster rotation (sin,cos)
    clusters_position = []

    for i in range(clusters):
        attempts = 100
        while attempts > 0:
            # prevent cluster position near borders (and on border)
            x = ((random() * 2.0 - 1.0) / (clusters + 1.0)) * clusters
            y = ((random() * 2.0 - 1.0) / (clusters + 1.0)) * clusters
            # ensure all clusters have a min separation to each other (search isn't optimized, not worth the effort)
            if all(
                ((cp[0][0] - x) * (cp[0][0] - x) + (cp[0][1] - y) * (cp[0][1] - y)) > (2.0 / clusters)
                for cp in clusters_position
            ):
                break
            attempts -= 1
        rotation = uniform(0.0, pi)
        clusters_position.append(((x, y), (sin(rotation), cos(rotation))))

    for i in range(size):
        attempts = 100
        while attempts > 0:
            if random() < system_noise:
                x = random() * 2.0 - 1.0
                y = random() * 2.0 - 1.0
            else:
                cluster = clusters_position[i % len(clusters_position)]
                radius = random()
                angle = uniform(0.0, 2.0 * pi)

                x1 = radius * cos(angle)
                y1 = radius * sin(angle) * ellipse_width_vs_height

                x = x1 * cluster[1][1] + y1 * cluster[1][0]
                y = (-x1) * cluster[1][0] + y1 * cluster[1][1]

                x = x / sqrt(float(clusters)) + cluster[0][0]
                y = y / sqrt(float(clusters)) + cluster[0][1]

            x = (x + 1) * width / 2.0
            y = (y + 1) * width / 2.0

            if (x < 0) or (width <= x) or (y < 0) or (width <= y):
                attempts -= 1
                continue

            # see if new star is too close to any existing star; if so, we try again
            if adjacency_grid.too_close_to_other_positions((x, y)):
                attempts -= 1
                continue

            # add the new star location
            pos = (x, y)
            adjacency_grid.insert_pos(pos)
            positions.append(pos)
            break

        if not attempts:
            print(
                f"Cluster galaxy shape: giving up on placing star {i}, can't "
                "find position sufficiently far from other systems"
            )


def ring_galaxy_calc_positions(positions, adjacency_grid, size, width):
    """
    Calculate positions for the ring galaxy shape.
    """
    ring_width = width / 4.0
    ring_radius = (width - ring_width) / 2.0

    for i in range(size):
        attempts = 100
        while attempts > 0:
            theta = uniform(0.0, 2.0 * pi)
            radius = gauss(ring_radius, ring_width / 3.0)

            x = width / 2.0 + radius * cos(theta)
            y = width / 2.0 + radius * sin(theta)

            if (x < 0) or (width <= x) or (y < 0) or (width <= y):
                attempts -= 1
                continue

            # see if new star is too close to any existing star; if so, we try again
            if adjacency_grid.too_close_to_other_positions((x, y)):
                attempts -= 1
                continue

            # add the new star location
            pos = (x, y)
            adjacency_grid.insert_pos(pos)
            positions.append(pos)
            break

        if not attempts:
            print(
                f"Ring galaxy shape: giving up on placing star {i}, can't"
                " find position sufficiently far from other systems"
            )


def box_galaxy_calc_positions(positions, adjacency_grid, size, width):
    """
    Calculate positions for the box galaxy shape.

    :param float width:
    """
    for i in range(size):
        attempts = 100
        while attempts > 0:
            x = width * random()
            y = width * random()

            if (x < 0) or (width <= x) or (y < 0) or (width <= y):
                attempts -= 1
                continue

            # see if new star is too close to any existing star; if so, we try again
            if adjacency_grid.too_close_to_other_positions((x, y)):
                attempts -= 1
                continue

            # add the new star location
            pos = (x, y)
            adjacency_grid.insert_pos(pos)
            positions.append(pos)
            break

        if not attempts:
            print(
                f"Box galaxy shape: giving up on placing star {i}, can't find"
                " position sufficiently far from other systems"
            )


def irregular_galaxy_calc_positions(positions, adjacency_grid, size, width):
    """
    Calculate positions for the irregular galaxy shape.
    """
    max_delta = max(min(float(universe_tables.MAX_STARLANE_LENGTH), width / 10.0), adjacency_grid.min_dist * 2.0)
    print(f"Irregular galaxy shape: max delta distance = {max_delta}")
    origin_x, origin_y = width / 2.0, width / 2.0
    prev_x, prev_y = origin_x, origin_y
    reset_to_origin = 0
    for _ in range(size):
        attempts = 100
        found = False
        while (attempts > 0) and not found:
            attempts -= 1
            x = prev_x + uniform(-max_delta, max_delta)
            y = prev_y + uniform(-max_delta, max_delta)
            if util.distance((x, y), (origin_x, origin_y)) > width * 0.45:
                prev_x, prev_y = origin_x, origin_y
                reset_to_origin += 1
                continue
            found = not adjacency_grid.too_close_to_other_positions((x, y))
            if attempts % 10:
                prev_x, prev_y = x, y
        if found:
            pos = (x, y)
            adjacency_grid.insert_pos(pos)
            positions.append(pos)
        prev_x, prev_y = x, y
    print(f"Reset to origin {reset_to_origin} times")


def recalc_universe_width(positions):
    """
    Recalculates the universe width. This is done by shifting all positions by a delta so too much "extra space"
    beyond the uppermost, lowermost, leftmost and rightmost positions is cropped, and adjust the universe width
    accordingly.

    Returns the new universe width and the recalculated positions.
    """
    print("Recalculating universe width...")
    # first, get the uppermost, lowermost, leftmost and rightmost positions
    # (these are those with their x or y coordinate closest to or farthest away from the x or y axis)
    min_x = min(positions, key=lambda p: p[0])[0]
    min_y = min(positions, key=lambda p: p[1])[1]
    max_x = max(positions, key=lambda p: p[0])[0]
    max_y = max(positions, key=lambda p: p[1])[1]
    print(f"...the leftmost system position is at x coordinate {min_x}")
    print(f"...the uppermost system position is at y coordinate {min_y}")
    print(f"...the rightmost system position is at x coordinate {max_x}")
    print(f"...the lowermost system position is at y coordinate {max_y}")

    # calculate the actual universe width by determining the width and height of an rectangle that encompasses all
    # positions, and take the greater of the two as the new actual width for the universe
    # also add a constant value to the width so we have some small space around
    width = max_x - min_x
    height = max_y - min_y
    actual_width = max(width, height) + 20.0
    print(f"...recalculated universe width: {actual_width}")

    # shift all positions so the entire map is centered in a quadratic box of the width we just calculated
    # this box defines the extends of our universe
    delta_x = ((actual_width - width) / 2) - min_x
    delta_y = ((actual_width - height) / 2) - min_y
    print(f"...shifting all system positions by {delta_x}/{delta_y}")
    new_positions = [(p[0] + delta_x, p[1] + delta_y) for p in positions]

    print(f"...the leftmost system position is now at x coordinate {min(new_positions, key=lambda p: p[0])[0]}")
    print(f"...the uppermost system position is now at y coordinate {min(new_positions, key=lambda p: p[1])[1]}")
    print(f"...the rightmost system position is now at x coordinate {max(new_positions, key=lambda p: p[0])[0]}")
    print(f"...the lowermost system position is now at y coordinate {max(new_positions, key=lambda p: p[1])[1]}")

    return actual_width, new_positions


def calc_star_system_positions(gsd):
    """
    Calculates list of positions (x, y) for a given galaxy shape,
    number of systems and width
    Uses universe generator helper functions provided by the API
    """

    # calculate typical width for universe based on number of systems
    width = calc_universe_width(gsd.shape, gsd.size)
    print(f"Set universe width to {width}")
    fo.set_universe_width(width)

    positions = []
    adjacency_grid = AdjacencyGrid(width)

    print(f"Creating {gsd.shape} galaxy shape")
    if gsd.shape == fo.galaxyShape.spiral2:
        spiral_galaxy_calc_positions(positions, adjacency_grid, 2, gsd.size, width)
    elif gsd.shape == fo.galaxyShape.spiral3:
        spiral_galaxy_calc_positions(positions, adjacency_grid, 3, gsd.size, width)
    elif gsd.shape == fo.galaxyShape.spiral4:
        spiral_galaxy_calc_positions(positions, adjacency_grid, 4, gsd.size, width)
    elif gsd.shape == fo.galaxyShape.elliptical:
        elliptical_galaxy_calc_positions(positions, adjacency_grid, gsd.size, width)
    elif gsd.shape == fo.galaxyShape.disc:
        disc_galaxy_calc_positions(positions, adjacency_grid, gsd.size, width)
    elif gsd.shape == fo.galaxyShape.cluster:
        cluster_galaxy_calc_positions(positions, adjacency_grid, gsd.size, width)
    elif gsd.shape == fo.galaxyShape.ring:
        ring_galaxy_calc_positions(positions, adjacency_grid, gsd.size, width)
    elif gsd.shape == fo.galaxyShape.irregular:
        irregular_galaxy_calc_positions(positions, adjacency_grid, gsd.size, width)

    # Check if any positions have been calculated...
    if not positions:
        # ...if not, fall back on box shape
        box_galaxy_calc_positions(positions, adjacency_grid, gsd.size, width)

    enforce_max_distance(positions, adjacency_grid)

    # to avoid having too much "extra space" around the system positions of our galaxy map, recalculate the universe
    # width and shift all positions accordingly
    width, positions = recalc_universe_width(positions)
    print(f"Set universe width to {width}")
    fo.set_universe_width(width)

    return positions



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="PRO_NDIM_ASSMB",
    description="PRO_NDIM_ASSMB_DESC",
    short_description="BUILDING_UNLOCK_SHORT_DESC",
    category="PRODUCTION_CATEGORY",
    researchcost=320 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_PRODUCTION_CATEGORY"],
    prerequisites=["LRN_NDIM_SUBSPACE"],
    unlock=Item(type=UnlockBuilding, name="BLD_HYPER_DAM"),
    graphic="icons/tech/n-dimensional_assembly.png",
)



from focs._effects import (
    AddSpecial,
    Conditional,
    ContainedBy,
    Contains,
    DesignHasPart,
    Destroy,
    DirectDistanceBetween,
    EffectsGroup,
    EnemyOf,
    Fleet,
    Focus,
    GenerateSitRepMessage,
    InSystem,
    IsBuilding,
    IsSource,
    LocalCandidate,
    MoveTo,
    MoveTowards,
    Number,
    Object,
    OwnedBy,
    Planet,
    Random,
    RootCandidate,
    SetMaxSupply,
    SetPopulation,
    SetStealth,
    SetTargetPopulation,
    Ship,
    Source,
    Stationary,
    System,
    Target,
    Turn,
    Unowned,
    Value,
    WithinDistance,
    WithinStarlaneJumps,
)
from macros.priorities import (
    POPULATION_DEFAULT_PRIORITY,
    POPULATION_FIRST_PRIORITY,
    TARGET_POPULATION_AFTER_SCALING_PRIORITY,
)

PLANETARY_DRIVE_ACTIVATION = (
    Planet()
    & Focus(type=["FOCUS_PLANET_DRIVE"])
    & WithinStarlaneJumps(
        jumps=1,
        condition=System
        & Contains(
            (
                IsBuilding(name=["BLD_PLANET_BEACON"])
                | (
                    Ship
                    & DesignHasPart(low=1, high=999, name="SP_PLANET_BEACON")
                    & Turn(low=LocalCandidate.ArrivedOnTurn + 1)
                )
            )
            & OwnedBy(empire=Source.Owner)
        )
        & ~Contains(IsSource),
    )
)

ADVANCED_FOCUS_EFFECTS = [
    EffectsGroup(
        scope=IsSource, activation=Planet() & Focus(type=["FOCUS_STEALTH"]), effects=SetStealth(value=Value + 15)
    ),
    EffectsGroup(
        scope=Planet()
        & OwnedBy(affiliation=EnemyOf, empire=Source.Owner)
        & WithinStarlaneJumps(jumps=4, condition=IsSource)
        & ~Number(low=1, condition=IsBuilding(name=["BLD_GENOME_BANK"]) & OwnedBy(empire=RootCandidate.Owner)),
        activation=Focus(type=["FOCUS_BIOTERROR"]),
        priority=TARGET_POPULATION_AFTER_SCALING_PRIORITY,
        effects=SetTargetPopulation(value=Value - 4),
    ),
    EffectsGroup(
        scope=Fleet
        & OwnedBy(empire=Source.Owner)
        & ContainedBy(System & Contains(Planet() & OwnedBy(empire=Source.Owner) & Focus(type=["FOCUS_STARGATE_SEND"]))),
        activation=Planet() & Focus(type=["FOCUS_STARGATE_RECEIVE"]),
        stackinggroup="STARGATE_STACK",
        effects=[
            GenerateSitRepMessage(
                message="EFFECT_STARGATE",
                label="EFFECT_STARGATE_LABEL",
                icon="icons/focus/stargate_receive.png",
                parameters={"fleet": Target.ID, "system": Source.SystemID},
                empire=Source.Owner,
            ),
            MoveTo(destination=Contains(IsSource) & System),
        ],
    ),
    EffectsGroup(
        scope=IsSource,
        activation=PLANETARY_DRIVE_ACTIVATION,
        # add the special with higher priority, so it can trigger with the same
        # priotity as the effect below
        priority=POPULATION_FIRST_PRIORITY,
        effects=AddSpecial(name="STARLANE_DRIVE_ACTIVATED_SPECIAL"),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=PLANETARY_DRIVE_ACTIVATION,
        priority=POPULATION_DEFAULT_PRIORITY,
        effects=[
            MoveTo(
                destination=System
                & WithinStarlaneJumps(jumps=1, condition=IsSource)
                & Contains(
                    (
                        IsBuilding(name=["BLD_PLANET_BEACON"])
                        | (
                            Ship
                            & DesignHasPart(low=1, high=999, name="SP_PLANET_BEACON")
                            & Turn(low=LocalCandidate.ArrivedOnTurn + 1)
                        )
                    )
                    & OwnedBy(empire=Source.Owner)
                )
                & ~Contains(IsSource)
            ),
            GenerateSitRepMessage(
                message="EFFECT_PLANET_DRIVE",
                label="EFFECT_PLANET_DRIVE_LABEL",
                icon="icons/building/planetary_stardrive.png",
                parameters={"planet": Source.ID, "system": Source.SystemID},
                empire=Source.Owner,
            ),
            SetPopulation(value=Value / 2),
        ],
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Random(probability=0.5)
        & Planet()
        & Focus(type=["FOCUS_PLANET_DRIVE"])
        & WithinStarlaneJumps(
            jumps=1,
            condition=System
            & Contains(
                (
                    IsBuilding(name=["BLD_PLANET_BEACON"])
                    | (
                        Ship
                        & DesignHasPart(low=1, high=999, name="SP_PLANET_BEACON")
                        & Turn(low=LocalCandidate.ArrivedOnTurn + 1)
                    )
                )
                & OwnedBy(empire=Source.Owner)
            )
            & ~Contains(IsSource),
        )
        & ~WithinDistance(distance=200, condition=IsBuilding(name=["BLD_LIGHTHOUSE"])),
        effects=[
            GenerateSitRepMessage(
                message="SITREP_PLANET_DRIVE_FAILURE",
                label="SITREP_PLANET_DRIVE_FAILURE_LABEL",
                icon="icons/sitrep/colony_destroyed.png",
                parameters={"planet": Source.ID, "system": Source.SystemID},
                empire=Source.Owner,
            ),
            Destroy,
        ],
    ),
    EffectsGroup(
        scope=Fleet
        & ~Stationary
        & ~InSystem()
        & (OwnedBy(affiliation=EnemyOf, empire=Source.Owner) | Unowned)
        & (Source.System.ID == LocalCandidate.NextSystemID),
        activation=Planet() & Focus(type=["FOCUS_DISTORTION"]),
        stackinggroup="DISTORTION_MOVEMENT_STACK",
        effects=Conditional(
            condition=WithinDistance(distance=40, condition=Object(id=RootCandidate.PreviousSystemID)),
            effects=[
                GenerateSitRepMessage(
                    message="EFFECT_FLEET_MOVED_TO",
                    label="EFFECT_FLEET_MOVED_TO_LABEL",
                    parameters={
                        "fleet": Target.ID,
                        "system": Target.PreviousSystemID,
                        "rawtext": DirectDistanceBetween(Target.ID, Target.PreviousSystemID),
                        "planet": Source.ID,
                    },
                    empire=Source.Owner,
                ),
                MoveTo(destination=Object(id=Target.PreviousSystemID)),
            ],
            else_=[
                GenerateSitRepMessage(
                    message="EFFECT_FLEET_MOVED_TOWARDS",
                    label="EFFECT_FLEET_MOVED_TOWARDS_LABEL",
                    parameters={
                        "fleet": Target.ID,
                        "system": Target.PreviousSystemID,
                        "rawtext": DirectDistanceBetween(Target.ID, Target.PreviousSystemID),
                        "planet": Source.ID,
                    },
                    empire=Source.Owner,
                ),
                MoveTowards(speed=40, target=Object(id=Target.PreviousSystemID)),
            ],
        ),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Focus(type=["FOCUS_LOGISTICS"]),
        accountinglabel="SHP_INTSTEL_LOG",
        effects=SetMaxSupply(value=Value + 3),
    ),
]



# In typical usage this is used as a plain macro, not a substitution macro; the substitution is instead done within the calling macro
from focs._effects import DesignHasPart, LocalCandidate, OwnedBy, Ship, Source, TurnTechResearched


def EMPIRE_OWNED_SHIP_WITH_PART(part_name: str):
    return Ship & OwnedBy(empire=Source.Owner) & DesignHasPart(name=part_name)


# For inclusion in a scope macro for an upgrade increase to a Max type meter (MaxCapacity or MaxSecondaryStat), and that scope
# should have already limited scope to ships owned by an empire
# In most cases the substitution provided for @1@ may simply be "CurrentContent" (without the quote marks) but
# if providing an explicit tech name then include quotes with the substitution,
# i.e., either use
# SHIP_PART_UPGRADE_RESUPPLY_CHECK("SOME_TECH")
# or
# SHIP_PART_UPGRADE_RESUPPLY_CHECK(CurrentContent)
def SHIP_PART_UPGRADE_RESUPPLY_CHECK(tech_name: str):
    return LocalCandidate.LastTurnResupplied >= TurnTechResearched(empire=LocalCandidate.Owner, name=tech_name)


ARBITRARY_BIG_NUMBER_FOR_METER_TOPUP = 1000000



from focs._types import _Condition, _EffectGroup

class _DestroyAction: ...

DestroyOnCapture = _DestroyAction()

def BuildingType(
    *,
    name: str,
    description: str,
    buildcost: int | float,
    buildtime: int,
    location: _Condition,
    effectsgroups: list[_EffectGroup],
    icon: str,
    captureresult=_DestroyAction,
    enqueuelocation=_Condition(),
): ...
